using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data;
using System.Data.Common;
using System.Globalization;
using System.Reflection;
using System.Text;
using Produmex.Foundation;
using Produmex.Foundation.Barcode;
using Produmex.Foundation.Convertors;
using Produmex.Foundation.Data.Sbo;
using Produmex.Foundation.Data.Sbo.BusinessObjects;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Convertors;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions.Tables;
using Produmex.Foundation.Data.Sbo.Providers;
using Produmex.Foundation.Data.Sbo.Utilities;
using Produmex.Foundation.Data.SqlClient;
using Produmex.Foundation.Diagnostics;
using Produmex.Foundation.GS1;
using Produmex.Foundation.Messages;
using Produmex.Foundation.SlimScreen;
using Produmex.Foundation.SlimScreen.Interfaces;
using Produmex.Foundation.SlimScreen.Interfaces.Definitions;
using Produmex.Foundation.SlimScreen.Interfaces.Definitions.KnownDataSets;
using Produmex.Foundation.SlimScreen.WinGui;
using Produmex.Foundation.Workflows;
using Produmex.Foundation.Workflows.Parameters;
using Produmex.Foundation.Wwf.Sbo.DataObjects;
using Produmex.Foundation.Wwf.Sbo.LocalServices;
using Produmex.Sbo.Logex.Data;
using Produmex.Sbo.Logex.Data.BusinessObjects;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.Data.DataObjects;
using Produmex.Sbo.Logex.Data.Providers;
using Produmex.Foundation.Data;
using Produmex.Foundation.Data.DbClient;

namespace Produmex.Sbo.Logex.WorkflowScripts
{
    /// <summary>
    /// SalesReturn
    /// </summary>
    public class WorkflowScript_TeoxaneSalesReturnScript : WorkflowInstanceScriptBase
    {
        /// <summary>
        /// Definitions
        /// </summary>
        private static class Definitions
        {
            /// <summary>
            /// The separator for separating strings values
            /// </summary>
            public const string SEPARATOR = ", ";
        }


        // Input parameters
        public ReadOnlyBinder<CultureInfo> DefaultCultureInfo;
        public ReadOnlyBinder<PmxOseCompany> PmxOseCompany;

        private CultureInfo current_DefaultCultureInfo
        {
            get { return DefaultCultureInfo.Get(); }
        }

        private PmxOseCompany currentPmxOseCompany
        {
            get { return PmxOseCompany.Get(); }
        }



        //Subflows
        WorkflowInstanceBase m_enterQuantitySubFlow;

        protected bool m_subFlowBackRequested;
        protected Collection<string> m_subFlowListOfSerialNumbers;
        protected double m_subFlowQuantity;

        #region TEOXANE

        // Table definition for PMX_TEOX_SERNUM
        private class PmxTeoxaneSerialNumbersTable
        {
            public const string NAME = "PMX_TEOX_SERNUM";

            public class Columns
            {
                public static class BatchNumber
                {
                    public const string NAME = "LotNr";
                }

                public static class FirstSeqNr
                {
                    public const string NAME = "FirstSeqNr";
                }

                public static class LastSeqNr
                {
                    public const string NAME = "LastSeqNr";
                }

                public static class Quantity
                {
                    public const string NAME = "Quantity";
                }

                public static class PickListDocEntry
                {
                    public const string NAME = "PickListDocEntry";
                }

                public static class PickListLineNum
                {
                    public const string NAME = "PickListLineNum";
                }

            }
        }

        // Table definition for PMX_TEOX_SERNUM_LOG
        private class PmxTeoxaneSerialNumbersLogTable
        {
            public const string NAME = "PMX_TEOX_SERNUM_LOG";

            public class Columns
            {
                public static class InternalKey
                {
                    public const string NAME = "InternalKey";
                }

                public static class BatchNumber
                {
                    public const string NAME = "LotNr";
                }

                public static class FirstSeqNr
                {
                    public const string NAME = "FirstSeqNr";
                }

                public static class LastSeqNr
                {
                    public const string NAME = "LastSeqNr";
                }

                public static class Quantity
                {
                    public const string NAME = "Quantity";
                }

                public static class PickListDocEntry
                {
                    public const string NAME = "PickListDocEntry";
                }

                public static class PickListLineNum
                {
                    public const string NAME = "PickListLineNum";
                }

                public static class SalesReturnDocEntry
                {
                    public const string NAME = "SalesReturnDocEntry";
                }

                public static class SalesReturnLineNum
                {
                    public const string NAME = "SalesReturnLineNum";
                }

            }
        }

        // Action object holding data and code for processing within transaction used for PickItem(...)
        private class ActionObject : IPreAndPostActionInTransaction
        {
            private class RangeQty
            {
                public int FirstSeqNr;
                public int LastSeqNr;
                public int Quantity;

                public RangeQty( int firstSeqNr, int lastSeqNr, int quantity )
                {
                    this.FirstSeqNr = firstSeqNr;
                    this.LastSeqNr = lastSeqNr;
                    this.Quantity = quantity;
                }
            }
            private List<RangeQty> m_rangeQtyBoxes = new List<RangeQty>();  // List of scanned boxes
            private List<RangeQty> m_rangeQtyPacks = new List<RangeQty>();  // List of scanned packs

            public SalesReturnItemInfo SalesReturnItemInfo;
            public int PickListDocEntry;
            public int PickListLineNum;
            public int SalesReturnDocEntry;
            public int? SalesReturnLineNum;

            /// <summary>
            /// Check if given range is free (not yet in list)
            /// </summary>
            /// <param name="firstSeqNr">The first seq nr.</param>
            /// <param name="lastSeqNr">The last seq nr.</param>
            /// <returns></returns>
            public bool RangeIsFree( int firstSeqNr, int lastSeqNr )
            {
                foreach ( RangeQty rq in m_rangeQtyBoxes )
                {
                    if ( rq.FirstSeqNr > lastSeqNr ) break;  // Current range past test range
                    if ( rq.LastSeqNr >= firstSeqNr ) return false;  // Current range overlaps test range, not free!
                }
                foreach ( RangeQty rq in m_rangeQtyPacks )
                {
                    if ( rq.FirstSeqNr > lastSeqNr ) break;  // Current range past test range
                    if ( rq.LastSeqNr >= firstSeqNr ) return false;  // Current range overlaps test range, not free!
                }
                return true;
            }

            /// <summary>
            /// Adds the range. Boxes are added separatly, pack serial numbers are joined with adjoining existing ranges
            /// </summary>
            /// <param name="firstSeqNr">The first seq nr.</param>
            /// <param name="lastSeqNr">The last seq nr.</param>
            /// <param name="quantity">The quantity.</param>
            public void AddRange( int firstSeqNr, int lastSeqNr, int quantity )
            {
                if ( firstSeqNr != lastSeqNr )
                {
                    // Add box in numerical order of serial number
                    for ( int i = 0; i < m_rangeQtyBoxes.Count; i++ )
                    {
                        // Look for the first existing range (box) with a first-seq-nr past our last-seq-nr
                        if ( m_rangeQtyBoxes[ i ].FirstSeqNr > lastSeqNr )
                        {
                            // Past the new range, insert new range (box) before this range (box)
                            m_rangeQtyBoxes.Insert( i, new RangeQty( firstSeqNr, lastSeqNr, quantity ) );
                            return;
                        }
                    }

                    // Add as first range in list or at the back of the list
                    m_rangeQtyBoxes.Add( new RangeQty( firstSeqNr, lastSeqNr, quantity ) );
                    return;
                }

                // Add single pack in numerical order of serial number
                // Join adjoinging ranges of serial numbers
                for ( int i = 0; i < m_rangeQtyPacks.Count; i++ )
                {
                    if ( m_rangeQtyPacks[ i ].LastSeqNr + 1 == firstSeqNr )  // Range ends where our new range begins
                    {
                        // Join new range at end of pointed range
                        m_rangeQtyPacks[ i ].LastSeqNr = lastSeqNr;
                        m_rangeQtyPacks[ i ].Quantity += quantity;

                        if ( i + 1 < m_rangeQtyPacks.Count )
                        {
                            // Join new range at begin of next range as well
                            if ( m_rangeQtyPacks[ i + 1 ].FirstSeqNr == lastSeqNr + 1 )
                            {
                                m_rangeQtyPacks[ i ].LastSeqNr = m_rangeQtyPacks[ i + 1 ].LastSeqNr;
                                m_rangeQtyPacks[ i ].Quantity += m_rangeQtyPacks[ i + 1 ].Quantity;
                                m_rangeQtyPacks.RemoveAt( i + 1 );
                            }
                        }

                        return;
                    }

                    if ( m_rangeQtyPacks[ i ].FirstSeqNr == lastSeqNr + 1 )  // Range begins where our new range ends
                    {
                        // Join new range at begin of pointed range
                        m_rangeQtyPacks[ i ].FirstSeqNr = firstSeqNr;
                        m_rangeQtyPacks[ i ].Quantity += quantity;
                        return;
                    }

                    if ( m_rangeQtyPacks[ i ].FirstSeqNr > lastSeqNr + 1 )
                    {
                        // Past the new range, so add as new range
                        m_rangeQtyPacks.Insert( i, new RangeQty( firstSeqNr, lastSeqNr, quantity ) );
                        return;
                    }
                }

                // Add as first or at the back of the list
                m_rangeQtyPacks.Add( new RangeQty( firstSeqNr, lastSeqNr, quantity ) );
            }

            /// <summary>
            /// Reduces a range in the database -> remove/extract a range from an existing range
            /// If the range to remove:
            ///  * falls within the existing range, do nothing, we cannot split a range in the database
            ///  * overlaps with the front of the existing range, cut from the front of the existing range
            ///  * overlaps with the back of the existing range, cut from the back of the existing range
            /// Cutting off of front or back also includes earlier ranges that would have required a 'split' (first case in list above)
            /// </summary>
            /// <param name="firstSeqNr">The first seq nr.</param>
            /// <param name="lastSeqNr">The last seq nr.</param>
            /// <param name="rowQty">The row qty.</param>
            /// <param name="removeFirstSeqNr">The remove first seq nr.</param>
            /// <param name="removeLastSeqNr">The remove last seq nr.</param>
            /// <param name="dbConn">The db conn.</param>
            /// <returns></returns>
            private int ReduceRange( int firstSeqNr, int lastSeqNr, int rowQty, int removeFirstSeqNr, int removeLastSeqNr, Produmex.Foundation.Data.Sbo.PmxDbConnection dbConn )
            {
                if ( removeFirstSeqNr > firstSeqNr && removeLastSeqNr < lastSeqNr )
                {
                    // Removing inner part, do nothing, we cannot split (which quantity for which part in case of a range with missing serial numbers)
                    return removeLastSeqNr - removeFirstSeqNr + 1;
                }

                if ( removeLastSeqNr < lastSeqNr )  // Removing front part
                {
                    int newFirstSeqNr = removeLastSeqNr + 1;
                    int overlappedQty = newFirstSeqNr - firstSeqNr;
                    int reduceQty = overlappedQty;

                    // Scan remaining range for already existing range assignments and remove them as well 
                    string query = TeoxaneBuildQuery.GetSerialNumberData( this.SalesReturnItemInfo.BatchNumber, removeLastSeqNr, lastSeqNr, true );
                    using ( ISboRecordset rs = SboRecordsetHelper.RunQuery( s_log, query, dbConn ) )
                    {
                        while ( !rs.EoF )
                        {
                            int currentFirstSeqNr = Convert.ToInt32( rs[ PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ] );
                            if ( currentFirstSeqNr > newFirstSeqNr ) break;  // Beyond join/overlap

                            int currentLastSeqNr = Convert.ToInt32( rs[ PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME ] );
                            if ( currentLastSeqNr >= newFirstSeqNr
                                && ( currentFirstSeqNr != firstSeqNr || currentLastSeqNr != lastSeqNr ) )
                            {
                                // Skip over adjoining range
                                newFirstSeqNr = Math.Min( currentLastSeqNr + 1, lastSeqNr + 1 );
                                reduceQty += newFirstSeqNr - currentFirstSeqNr;
                            }

                            rs.MoveNext();
                        }
                    }

                    if ( newFirstSeqNr <= lastSeqNr )
                    {
                        query = TeoxaneBuildQuery.MoveResizeRange( this.SalesReturnItemInfo.BatchNumber, firstSeqNr, newFirstSeqNr, lastSeqNr, rowQty - reduceQty );
                    }
                    else
                    {
                        // Nothing left of db range, remove it completely
                        query = TeoxaneBuildQuery.DeleteRange( this.SalesReturnItemInfo.BatchNumber, firstSeqNr );
                    }
                    SboRecordsetHelper.RunQueryNoReturn( s_log, query, dbConn );

                    return overlappedQty;
                }
                else  // Removing back part
                {
                    int newLastSeqNr = removeFirstSeqNr - 1;
                    int overlappedQty = lastSeqNr - newLastSeqNr;
                    int reduceQty = overlappedQty;

                    // Scan remaining range for already existing range assignments and remove them as well
                    string query = TeoxaneBuildQuery.GetSerialNumberData( this.SalesReturnItemInfo.BatchNumber, firstSeqNr, removeFirstSeqNr, true );
                    using ( ISboRecordset rs = SboRecordsetHelper.RunQuery( s_log, query, dbConn ) )
                    {
                        rs.MoveLast();
                        while ( !rs.BoF )
                        {
                            int currentLastSeqNr = Convert.ToInt32( rs[ PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME ] );
                            if ( currentLastSeqNr < newLastSeqNr ) break;  // Beyond join/overlap

                            int currentFirstSeqNr = Convert.ToInt32( rs[ PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ] );
                            if ( currentFirstSeqNr <= newLastSeqNr
                                && ( currentFirstSeqNr != firstSeqNr || currentLastSeqNr != lastSeqNr ) )
                            {
                                // Skip over adjoining range
                                newLastSeqNr = Math.Max( firstSeqNr - 1, currentFirstSeqNr - 1 );
                                reduceQty += currentLastSeqNr - newLastSeqNr;
                            }

                            rs.MovePrevious();
                        }
                    }

                    if ( firstSeqNr <= newLastSeqNr )
                    {
                        query = TeoxaneBuildQuery.MoveResizeRange( this.SalesReturnItemInfo.BatchNumber, firstSeqNr, firstSeqNr, newLastSeqNr, rowQty - reduceQty );
                    }
                    else
                    {
                        // Nothing left of db range, remove it completely
                        query = TeoxaneBuildQuery.DeleteRange( this.SalesReturnItemInfo.BatchNumber, firstSeqNr );
                    }
                    SboRecordsetHelper.RunQueryNoReturn( s_log, query, dbConn );

                    return overlappedQty;
                }
            }

            private void UpdateRange( RangeQty rangeQty, bool isBox, Produmex.Foundation.Data.Sbo.PmxDbConnection dbConn )
            {
                string query = TeoxaneBuildQuery.GetSerialNumberData( this.SalesReturnItemInfo.BatchNumber, rangeQty.FirstSeqNr, rangeQty.LastSeqNr, true );
                using ( ISboRecordset rs = SboRecordsetHelper.RunQuery( s_log, query, dbConn ) )
                {
                    if ( rs.EoF )
                    {
                        // No records found
                        throw new ProdumexException( "Serial number data unexpectedly changed, items not picked!" );
                    }

                    int pendingUpdateFirstSeqNr = -1;
                    int pendingUpdateNewLastSeqNr = -1;
                    int pendingUpdateNewQuantity = -1;
                    bool pendingUpdateIsInsert = false;
                    bool dbQtyIsExact = true;
                    bool reduceAtEnd = false;
                    bool reduceAtBegin = false;

                    int dbQty = 0;
                    int lastBigRangeFirstSeqNr = -1;
                    int lastBigRangeLastSeqNr = -1;
                    while ( !rs.EoF )
                    {
                        //if ( SystemInt32Convertor.ObjectToIntOrNull( rs[ TeoxaneDefs.ActivePickListDocEntry ] ).HasValue )
                        //{
                        //    // Serial number already used
                        //    throw new ProdumexException( "Serial number data unexpectedly changed, items not picked!" );
                        //}

                        int rowQty = rs.GetTypedValue<int>( PmxTeoxaneSerialNumbersTable.Columns.Quantity.NAME );
                        if ( rowQty > 0 )
                        {
                            int currentFirstSeqNr = rs.GetTypedValue<int>( PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME );
                            int currentLastSeqNr = rs.GetTypedValue<int>( PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME );

                            if ( currentFirstSeqNr >= lastBigRangeFirstSeqNr && currentLastSeqNr <= lastBigRangeLastSeqNr )
                            {
                                // Exists inside bigger range, delete it (single packs scanned and then picklist closed results in remnants)
                                query = TeoxaneBuildQuery.DeleteRange( this.SalesReturnItemInfo.BatchNumber, currentFirstSeqNr );
                                SboRecordsetHelper.RunQueryNoReturn( s_log, query, dbConn );
                            }
                            else
                            {
                                lastBigRangeFirstSeqNr = currentFirstSeqNr;
                                lastBigRangeLastSeqNr = currentLastSeqNr;

                                bool reduceDone = false;

                                if ( currentFirstSeqNr < rangeQty.FirstSeqNr )  // Current range begins before wanted range
                                {
                                    if ( reduceAtEnd )
                                    {
                                        // This can only occur ones, otherwise the db range data is corrupt (overlapping ranges)
                                        throw new ProdumexException( "Serial number data in database is incorrect, items not picked!" );
                                    }
                                    reduceAtEnd = true;
                                    reduceDone = true;

                                    if ( currentLastSeqNr - currentFirstSeqNr + 1 != rowQty ) dbQtyIsExact = false;

                                    // Reduce current range
                                    int overlappedQty = ReduceRange( currentFirstSeqNr, currentLastSeqNr, rowQty,
                                        rangeQty.FirstSeqNr, rangeQty.LastSeqNr, dbConn );

                                    currentFirstSeqNr = rangeQty.FirstSeqNr;
                                    currentLastSeqNr = Math.Min( currentLastSeqNr, rangeQty.LastSeqNr );
                                    if ( pendingUpdateFirstSeqNr < 0 ) pendingUpdateIsInsert = true;

                                    rowQty = overlappedQty;
                                }
                                else if ( currentLastSeqNr > rangeQty.LastSeqNr )  // Current range ends beyond wanted range
                                {
                                    if ( reduceAtBegin )
                                    {
                                        // This can only occur ones, otherwise the db range data is corrupt (overlapping ranges)
                                        throw new ProdumexException( "Serial number data in database is incorrect, items not picked!" );
                                    }
                                    reduceAtBegin = true;
                                    reduceDone = true;

                                    if ( currentLastSeqNr - currentFirstSeqNr + 1 != rowQty ) dbQtyIsExact = false;

                                    // Reduce current range
                                    int overlappedQty = ReduceRange( currentFirstSeqNr, currentLastSeqNr, rowQty,
                                        rangeQty.FirstSeqNr, rangeQty.LastSeqNr, dbConn );

                                    currentLastSeqNr = rangeQty.LastSeqNr;
                                    if ( pendingUpdateFirstSeqNr < 0 ) pendingUpdateIsInsert = true;

                                    rowQty = overlappedQty;
                                }

                                if ( pendingUpdateFirstSeqNr >= 0 )
                                {
                                    // There is a pending update, join with it if adjoining range, otherwise execute pending update
                                    if ( pendingUpdateNewLastSeqNr + 1 != currentFirstSeqNr )
                                    {
                                        // Current range doesn't join with pending update, execute pending update
                                        if ( pendingUpdateIsInsert )
                                        {
                                            query = TeoxaneBuildQuery.InsertRange( this.SalesReturnItemInfo.BatchNumber, pendingUpdateFirstSeqNr, pendingUpdateNewLastSeqNr, pendingUpdateNewQuantity,
                                                null, null );
                                        }
                                        else
                                        {
                                            query = TeoxaneBuildQuery.UpdateRange( this.SalesReturnItemInfo.BatchNumber, pendingUpdateFirstSeqNr, pendingUpdateNewLastSeqNr, pendingUpdateNewQuantity,
                                                null, null );
                                        }
                                        SboRecordsetHelper.RunQueryNoReturn( s_log, query, dbConn );

                                        pendingUpdateFirstSeqNr = -1;
                                        pendingUpdateIsInsert = false;
                                    }
                                    else if ( !reduceDone )
                                    {
                                        // Current range does join with pending update, delete current range
                                        query = TeoxaneBuildQuery.DeleteRange( this.SalesReturnItemInfo.BatchNumber, currentFirstSeqNr );
                                        SboRecordsetHelper.RunQueryNoReturn( s_log, query, dbConn );
                                    }
                                }
                                if ( pendingUpdateFirstSeqNr < 0 )
                                {
                                    pendingUpdateFirstSeqNr = currentFirstSeqNr;
                                    pendingUpdateNewQuantity = 0;
                                }
                                pendingUpdateNewLastSeqNr = currentLastSeqNr;
                                pendingUpdateNewQuantity += rowQty;

                                dbQty += rowQty;
                            }
                        }
                        else
                        {
                            // Check here for voided serial numbers
                        }

                        rs.MoveNext();
                    }
                    if ( dbQty != rangeQty.Quantity )
                    {
                        // Quantity doesn't match anymore
                        throw new ProdumexException( "Serial number data unexpectedly changed, items not picked!" );
                    }
                    if ( pendingUpdateFirstSeqNr >= 0 )
                    {
                        // There is a pending update, execute it
                        if ( pendingUpdateIsInsert )
                        {
                            query = TeoxaneBuildQuery.InsertRange( this.SalesReturnItemInfo.BatchNumber, pendingUpdateFirstSeqNr, pendingUpdateNewLastSeqNr, pendingUpdateNewQuantity,
                                null, null );
                        }
                        else
                        {
                            query = TeoxaneBuildQuery.UpdateRange( this.SalesReturnItemInfo.BatchNumber, pendingUpdateFirstSeqNr, pendingUpdateNewLastSeqNr, pendingUpdateNewQuantity,
                                null, null );
                        }
                        SboRecordsetHelper.RunQueryNoReturn( s_log, query, dbConn );
                    }
                }

                query = TeoxaneBuildQuery.AddLog( this.SalesReturnItemInfo.BatchNumber, rangeQty.FirstSeqNr, rangeQty.LastSeqNr, rangeQty.Quantity,
                    this.PickListDocEntry, this.PickListLineNum, this.SalesReturnDocEntry, this.SalesReturnLineNum );
                SboRecordsetHelper.RunQueryNoReturn( s_log, query, dbConn );
            }

            #region IPreAndPostActionInTransaction Members

            public System.Type RequiredConnectionType
            {
                get { return null; }
            }

            public void PreAction( Produmex.Foundation.Data.Sbo.PmxDbConnection dbConn )
            {
                // Now verify and update serial number ranges in database with DocEntry/LineNum of PickList

                // Update boxes
                foreach ( RangeQty rq in m_rangeQtyBoxes )
                {
                    UpdateRange( rq, true, dbConn );
                }

                // Update packs
                foreach ( RangeQty rq in m_rangeQtyPacks )
                {
                    UpdateRange( rq, false, dbConn );
                }
            }

            public void PostAction( Produmex.Foundation.Data.Sbo.PmxDbConnection dbConn )
            {
                // Do nothing
            }

            #endregion
        }

        private struct TeoxaneDefs
        {
            public const int PacksInBox = 100;

            public const string ActivePickListDocEntry = "ActivePickListDocEntry";
        }

        static class TeoxaneBuildQuery
        {
            internal static string GetSalesReturnLinesBaseInfo( int salesReturnDocEntry )
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat( "SELECT {0}",
                    SqlCommandHelper.EscapeTableColumnName( PmxSalesReturnLineTable.NAME, PmxSalesReturnLineTable.Columns.LineNum.NAME ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.EscapeTableColumnName( PmxSalesReturnLineTable.NAME, PmxSalesReturnLineTable.Columns.BaseType.NAME ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.EscapeTableColumnName( PmxSalesReturnLineTable.NAME, PmxSalesReturnLineTable.Columns.BaseEntry.NAME ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.EscapeTableColumnName( PmxSalesReturnLineTable.NAME, PmxSalesReturnLineTable.Columns.BaseLine.NAME ) );

                strBuilder.AppendFormat( "\r\n FROM {0}",
                    SqlCommandHelper.EscapeTableName( PmxSalesReturnLineTable.NAME ) );

                strBuilder.AppendFormat( "\r\n WHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxSalesReturnLineTable.NAME, PmxSalesReturnLineTable.Columns.DocEntry.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( salesReturnDocEntry, CultureInfo.InvariantCulture ) );

                return strBuilder.ToString();
            }

            internal static string GetSerialNumberData( string lotNr, int firstSeqNr, int lastSeqNr, bool updateLock )
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat( "SELECT {0}.*",
                    SqlCommandHelper.EscapeTableName( PmxTeoxaneSerialNumbersTable.NAME ) );
                strBuilder.Append( ", CASE" );
                {
                    strBuilder.AppendFormat( "\r\n    WHEN ( {0} <> {1}",
                        SqlCommandHelper.EscapeTableColumnName( PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.LineStatus.NAME ),
                        SqlCommandHelper.FormatAndEscapeColumnValue( PmxDocumentStatusColumnValues.Closed.VALUE, CultureInfo.InvariantCulture ) );
                    strBuilder.AppendFormat( " AND {0} <> {1} )",
                        SqlCommandHelper.EscapeTableColumnName( PmxPickListTable.NAME, PmxPickListTable.Columns.DocStatus.NAME ),
                        SqlCommandHelper.FormatAndEscapeColumnValue( PmxDocumentStatusColumnValues.Closed.VALUE, CultureInfo.InvariantCulture ) );
                    strBuilder.AppendFormat( " OR {0} IS NOT NULL",
                        SqlCommandHelper.EscapeTableColumnName( PmxLinkSalesDeliveryNoteLinePickListLineTable.NAME, PmxLinkSalesDeliveryNoteLinePickListLineTable.Columns.PmxSalesShippingDocEntry.NAME ) );
                    strBuilder.AppendFormat( " OR {0} IS NOT NULL",
                        SqlCommandHelper.EscapeTableColumnName( PmxLinkSalesDeliveryNoteLinePickListLineTable.NAME, PmxLinkSalesDeliveryNoteLinePickListLineTable.Columns.SalesDeliveryNoteLineDocEntry.NAME ) );
                    strBuilder.AppendFormat( " THEN {0}",
                        SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListDocEntry.NAME ) );
                }
                strBuilder.AppendFormat( "\r\n    ELSE NULL END AS {0}",
                    SqlCommandHelper.EscapeColumnName( TeoxaneDefs.ActivePickListDocEntry ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.EscapeTableColumnName( PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.ItemCode.NAME ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.EscapeTableColumnName( PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BatchNumber1.NAME ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.EscapeTableColumnName( PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BatchNumber2.NAME ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.EscapeTableColumnName( PmxLinkSalesDeliveryNoteLinePickListLineTable.NAME, PmxLinkSalesDeliveryNoteLinePickListLineTable.Columns.SalesDeliveryNoteLineDocEntry.NAME ) );

                strBuilder.AppendFormat( "\r\n FROM {0}",
                    SqlCommandHelper.EscapeTableName( PmxTeoxaneSerialNumbersTable.NAME ) );

                if ( updateLock )
                {
                    strBuilder.Append( " WITH ( UPDLOCK )" );
                }

                strBuilder.AppendFormat( "\r\n LEFT JOIN ( {0}",
                    SqlCommandHelper.EscapeTableName( PmxPickListLineTable.NAME ) );
                {
                    strBuilder.AppendFormat( "\r\n    INNER JOIN {0}",
                        SqlCommandHelper.EscapeTableName( PmxPickListTable.NAME ) );
                    strBuilder.AppendFormat( " ON {0} = {1}",
                        SqlCommandHelper.EscapeTableColumnName( PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME ),
                        SqlCommandHelper.EscapeTableColumnName( PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.DocEntry.NAME ) );
                }
                strBuilder.AppendFormat( "\r\n ) ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.DocEntry.NAME ),
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListDocEntry.NAME ) );
                strBuilder.AppendFormat( " AND {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.LineNum.NAME ),
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListLineNum.NAME ) );

                strBuilder.AppendFormat( "\r\n LEFT JOIN {0}",
                    SqlCommandHelper.EscapeTableName( PmxItemTransactionalInfoTable.NAME ) );
                strBuilder.AppendFormat( " ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.InternalKey.NAME ),
                    SqlCommandHelper.EscapeTableColumnName( PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.ItemTransactionalInfoKey.NAME ) );

                strBuilder.AppendFormat( "\r\n LEFT JOIN {0}",
                    SqlCommandHelper.EscapeTableName( PmxLinkSalesDeliveryNoteLinePickListLineTable.NAME ) );
                strBuilder.AppendFormat( " ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxLinkSalesDeliveryNoteLinePickListLineTable.NAME, PmxLinkSalesDeliveryNoteLinePickListLineTable.Columns.PickListLineDocEntry.NAME ),
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListDocEntry.NAME ) );
                strBuilder.AppendFormat( " AND {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxLinkSalesDeliveryNoteLinePickListLineTable.NAME, PmxLinkSalesDeliveryNoteLinePickListLineTable.Columns.PickListLineLineNum.NAME ),
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListLineNum.NAME ) );

                strBuilder.AppendFormat( "\r\n WHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.BatchNumber.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( lotNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( "\r\n AND {0} <= {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( lastSeqNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( "\r\n AND {0} >= {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( firstSeqNr, CultureInfo.InvariantCulture ) );

                strBuilder.AppendFormat( "\r\n ORDER BY {0}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ) );

                return strBuilder.ToString();
            }

            internal static string InsertRange( string lotNr, int firstSeqNr, int lastSeqNr, int quantity, int? pickListDocEntry, int? pickListLineNum )
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat( "INSERT INTO {0} ( ",
                    SqlCommandHelper.EscapeTableName( PmxTeoxaneSerialNumbersTable.NAME ) );

                strBuilder.Append( SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.BatchNumber.NAME ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.Quantity.NAME ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListDocEntry.NAME ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListLineNum.NAME ) );
                strBuilder.Append( " )" );

                strBuilder.Append( "\r\n VALUES ( " );
                strBuilder.Append( SqlCommandHelper.FormatAndEscapeColumnValue( lotNr, CultureInfo.InvariantCulture ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.FormatAndEscapeColumnValue( firstSeqNr, CultureInfo.InvariantCulture ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.FormatAndEscapeColumnValue( lastSeqNr, CultureInfo.InvariantCulture ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.FormatAndEscapeColumnValue( quantity, CultureInfo.InvariantCulture ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.FormatAndEscapeColumnValue( pickListDocEntry, CultureInfo.InvariantCulture ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.FormatAndEscapeColumnValue( pickListLineNum, CultureInfo.InvariantCulture ) );
                strBuilder.Append( " )" );

                return strBuilder.ToString();
            }

            internal static string UpdateRange( string lotNr, int firstSeqNr, int newLastSeqNr, int newQuantity, int? pickListDocEntry, int? pickListLineNum )
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat( "UPDATE {0}",
                    SqlCommandHelper.EscapeTableName( PmxTeoxaneSerialNumbersTable.NAME ) );

                strBuilder.AppendFormat( "\r\n SET {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( newLastSeqNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.Quantity.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( newQuantity, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListDocEntry.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( pickListDocEntry, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListLineNum.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( pickListLineNum, CultureInfo.InvariantCulture ) );

                strBuilder.AppendFormat( "\r\n WHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.BatchNumber.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( lotNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( "\r\n AND {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( firstSeqNr, CultureInfo.InvariantCulture ) );

                return strBuilder.ToString();
            }

            internal static string DeleteRange( string lotNr, int firstSeqNr )
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat( "DELETE FROM {0}",
                    SqlCommandHelper.EscapeTableName( PmxTeoxaneSerialNumbersTable.NAME ) );

                strBuilder.AppendFormat( "\r\n WHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.BatchNumber.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( lotNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( "\r\n AND {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( firstSeqNr, CultureInfo.InvariantCulture ) );

                return strBuilder.ToString();
            }

            internal static string MoveResizeRange( string lotNr, int firstSeqNr, int newFirstSeqNr, int newLastSeqNr, int newQuantity )
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat( "UPDATE {0}",
                    SqlCommandHelper.EscapeTableName( PmxTeoxaneSerialNumbersTable.NAME ) );

                strBuilder.AppendFormat( "\r\n SET {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( newFirstSeqNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( newLastSeqNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.Quantity.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( newQuantity, CultureInfo.InvariantCulture ) );

                strBuilder.AppendFormat( "\r\n WHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.BatchNumber.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( lotNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( "\r\n AND {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( firstSeqNr, CultureInfo.InvariantCulture ) );

                return strBuilder.ToString();
            }

            internal static string AddLog( string lotNr, int firstSeqNr, int lastSeqNr, int quantity, int? pickListDocEntry, int? pickListLineNum, int? salesReturnDocEntry, int? salesReturnLineNum )
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat( "INSERT INTO {0}",
                    SqlCommandHelper.EscapeTableName( PmxTeoxaneSerialNumbersLogTable.NAME ) );

                strBuilder.AppendFormat( " ( {0}",
                    SqlCommandHelper.EscapeColumnName( PmxTeoxaneSerialNumbersLogTable.Columns.BatchNumber.NAME ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.EscapeColumnName( PmxTeoxaneSerialNumbersLogTable.Columns.FirstSeqNr.NAME ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.EscapeColumnName( PmxTeoxaneSerialNumbersLogTable.Columns.LastSeqNr.NAME ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.EscapeColumnName( PmxTeoxaneSerialNumbersLogTable.Columns.Quantity.NAME ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.EscapeColumnName( PmxTeoxaneSerialNumbersLogTable.Columns.PickListDocEntry.NAME ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.EscapeColumnName( PmxTeoxaneSerialNumbersLogTable.Columns.PickListLineNum.NAME ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.EscapeColumnName( PmxTeoxaneSerialNumbersLogTable.Columns.SalesReturnDocEntry.NAME ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.EscapeColumnName( PmxTeoxaneSerialNumbersLogTable.Columns.SalesReturnLineNum.NAME ) );

                strBuilder.Append( " )\r\n VALUES" );

                strBuilder.AppendFormat( " ( {0}",
                    SqlCommandHelper.FormatAndEscapeColumnValue( lotNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.FormatAndEscapeColumnValue( firstSeqNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.FormatAndEscapeColumnValue( lastSeqNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.FormatAndEscapeColumnValue( quantity, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.FormatAndEscapeColumnValue( pickListDocEntry, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.FormatAndEscapeColumnValue( pickListLineNum, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.FormatAndEscapeColumnValue( salesReturnDocEntry, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0}",
                    SqlCommandHelper.FormatAndEscapeColumnValue( salesReturnLineNum, CultureInfo.InvariantCulture ) );

                strBuilder.Append( " )" );

                return strBuilder.ToString();
            }
        }

        #endregion

        ISboProviderService m_sboProviderService;



        private static readonly ILog s_log = LogProvider.GetLogger(MethodInfo.GetCurrentMethod().DeclaringType);


        #region Constructors

        public WorkflowScript_TeoxaneSalesReturnScript( WorkflowInstanceBase parent, WorkflowInstanceFactory factory )
            : base(parent, factory)
        {
            m_enterQuantitySubFlow = this.Create("EnterQuantityScript");

        }

        #endregion

        #region WorkflowInstanceScriptBase Members

        /// <summary>
        /// Executes this instance.
        /// </summary>
        protected override void Execute()
        {
            // Parameters in scope
            Session session = GetScopeParameter("Session") as Session;
            m_sboProviderService = GetScopeParameter("<WwfService>ISboService") as ISboProviderService;

            string initialErrorKey = null;
            string deviceID = session.DeviceInstanceId;
            SboUserInfo userInfo = m_sboProviderService.GetCurrentUser();
            Message msg = null;
            string query = null;
            PmxOseWarehouseInfo whsInfo = null;
            PmxItemInfo itemInfo = null;
            int numberOfReports = 0;
            bool goBack = false;
            ToolBarButtons toolBarButtons = ToolBarButtons.Back;
            Collection<string> listOfSerialNumbers = null;

            DataSet dsCustomers = null;
            string cardCode = null;
            string cardName = null;
            string[] buttonKeys = null;
            string filterButtonKey = null;
            Collection<SalesReturnItemInfo> salesReturnItemInfoList = null;
            List<ActionObject> teoxSerialNumbersList = null;
            DataSet dsDelivery = null;
            int? salesDeliveryDocEntry = null;
            string scanLabelScreenTitleKey = null;
            string[] allowedButtonsForLabelScreen = null;
            GS1Data gs1DataScanned = null;
            BarcodeContainer barcodeContainerObject = null;
            string buttonKey = null;

            double quantity = 0;
            string reasonCode = null;
            string batchNumber = null;
            DateTime? bestBeforeDate = null;
            string itemCode = null;
            string itemDescription = null;
            string itemName = null;
            string uom = null;
            string innerUom = null;
            double quantityPerUom = 1;
            Collection<OpenQuantityBySalesDeliveryLineNum> openQuantityList = null;
            string gtin = null;
            string sscc = null;
            DataSet dsItems = null;
            string batchNumber2 = null;
            DataSet dsBatchNumbers = null;
            int? itriKey = null;
            double batchQuantity = 0;
            DataSet ds = null;
            double maximumQuantity = 0;
            bool forceSerial = false;
            bool useRange = false;
            string docNum = null;
            DataSet dsReasons = null;

            int? salesReturnDocEntry = null;
            string locationCode = null;

            bool newSSCC = false;
            bool emptyLocations = false;
            int? dstLuid = null;
            string targetLogCarrierCode = null;
            string destinationPmxWhsCode = null;
			bool blockMoveWhenNotEmpty = false;
            string dstLocationCode = null;
            Collection<string> itemCodes = null;
            bool isLocked = false;
            Dictionary<string, PmxStorageLocationCanItemBeStoredInfo> canItemBeStoredInfoPerItemCode = null;

            whsInfo = m_sboProviderService.InvokeMethodWithDbConnection<PmxOseWarehouseInfo>(false, false, null, null, new DelegateWithPmxDbConnection<PmxOseWarehouseInfo>(ExecuteGetPmxWarehouseCode), deviceID);
            bool askReprint = false;

        Step_SelectCustomer:

            #region SELECT CUSTOMER

            // Get locations
            query = BuildQuery.GetCodeNameCustomersOrderByName();
            dsCustomers = m_sboProviderService.RunView(false, null, null, query);

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectCustomerScreen),
                DefaultCultureInfo.Get(), BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_SELECT_CUSTOMER",
                "CustomerDS", dsCustomers,
                "ShowStopButton", true
                ));

            msg = WaitForMessage();
            initialErrorKey = null;
            if (msg.Name.EndsWith(".CustomerSelected"))
            {
                cardCode = ExtractParameter<string>(msg.Parameters, "customerCode");
                goto Step_SelectFilter;
            }


            //back or stop requested
            return;

            #endregion

        Step_SelectFilter:

            #region SELECT FILTER

            salesDeliveryDocEntry = null;
            docNum = null;

            buttonKeys = new string[2] { "MSG_BUTTON_SELECT_DELIVERY", "MSG_BUTTON_NO_DELIVERY" };

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectButtonScreen),
                    DefaultCultureInfo.Get(), BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", "MSG_TITLE_SELECT_TASK",
                    "AllowedButtons", buttonKeys
                    ));
            msg = WaitForMessage();
            initialErrorKey = null;
            if (msg.Name.EndsWith(".ButtonClicked"))
            {
                //initialize list
                salesReturnItemInfoList = new Collection<SalesReturnItemInfo>();
                teoxSerialNumbersList = new List<ActionObject>();

                filterButtonKey = ExtractParameter<string>(msg.Parameters, "buttonKey");

                if (filterButtonKey == "MSG_BUTTON_SELECT_DELIVERY")
                {
                    goto Step_SelectDelivery;
                }

                //no link
                goto Step_StartScanLabel;
            }


            //Go back
            goto Step_SelectCustomer;
          

            #endregion

        Step_SelectDelivery:

            #region SELECT DELIVERY

            query = BuildQuery.GetSalesDeliveryNotesOrderByDate(cardCode);
            dsDelivery = m_sboProviderService.RunView(false, null, null, query);

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Retour.ISelectSalesDeliveryNoteScreen),
                   DefaultCultureInfo.Get(), BuildParamCollection(
                   "InitialErrorKey", initialErrorKey,
                   "TitleKey", "MSG_TITLE_SELECT_SALES_DELIVERY_NOTE",
                   "SalesDeliveryNoteDS", dsDelivery
                   ));
            msg = WaitForMessage();
            initialErrorKey = null;
            if (msg.Name.EndsWith(".SalesDeliveryNoteSelected"))
            {

                salesDeliveryDocEntry = ExtractParameter<int?>(msg.Parameters, "docEntry");

                DataRow[] rows = dsDelivery.Tables[0].Select(string.Format("{0}={1}", SalesDeliveryNoteDataSet.Columns.DOC_ENTRY.Name, salesDeliveryDocEntry.Value));

                if (rows.Length > 0)
                {
                    docNum = Convert.ToString(rows[0][SalesDeliveryNoteDataSet.Columns.DOC_NUM.Name]);
                }

                goto Step_StartScanLabel;
            }


            //Go back
            goto Step_SelectFilter;

            #endregion

        Step_StartScanLabel:

            #region START SCAN LABEL

            scanLabelScreenTitleKey = "MSG_TITLE_SCAN_GS1_OR_SERIAL_LABEL";
            if (salesReturnItemInfoList != null &&
                salesReturnItemInfoList.Count > 0)
            {
                allowedButtonsForLabelScreen = new string[2] { "MSG_BUTTON_NO_LABEL", "MSG_BUTTON_ADD_RETURN" };
            }
            else
            {
                allowedButtonsForLabelScreen = new string[1] { "MSG_BUTTON_NO_LABEL" };
            }

            gs1DataScanned = new GS1Data();
            barcodeContainerObject = null;

            //clear data
            quantity = 0;
            reasonCode = null;
            batchNumber = null;
            bestBeforeDate = null;
            itemCode = null;
            itemDescription = null;
            uom = null;
            innerUom = null;
            quantityPerUom = 1;
            openQuantityList = null;
            gtin = null;
            sscc = null;
            batchNumber2 = null;
            itriKey = null;
            batchQuantity = 0;

            string teoxLotNr = null;
            int teoxFirstSeqNr = -1;
            int teoxLastSeqNr = -1;
            int teoxQuantity = -1;
            string teoxBatchNum = null;
            string teoxBatchNum2 = null;
            int teoxPickListDocEntry = -1;
            int teoxPickListLineNum = -1;

            #endregion

        Step_ScanGS1Label:

            #region SCAN GS1 LABEL
              session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IIdentifyGS1DataScreen),
                DefaultCultureInfo.Get(), BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", scanLabelScreenTitleKey,
                    "ImageType", GS1ImageType.BoxImage,
                    "AllowedButtons", allowedButtonsForLabelScreen
                    ));

            msg = WaitForMessage();
            // Reset latest error
            initialErrorKey = null;
           
            if (msg.Name.EndsWith(".DataEntered"))
            {
                barcodeContainerObject = ExtractParameter<BarcodeContainer>(msg.Parameters, "data");

                if (this.IsSSCCOnBarcode(barcodeContainerObject.GS1DataObject))
                {
                    sscc = barcodeContainerObject.GS1DataObject.Sscc;
                    goto Step_GetProducts;
                }

                if (this.IsProductOnBarcode(barcodeContainerObject))
                {
                    goto Step_GetProductFromBarCode;
                }
                

                //not all data, so scan again
                scanLabelScreenTitleKey = "MSG_TITLE_SCAN_NEXT_GS1_LABEL";
                if (salesReturnItemInfoList != null &&
                        salesReturnItemInfoList.Count > 0)
                {
                    allowedButtonsForLabelScreen = new string[2] { "MSG_BUTTON_DONE", "MSG_BUTTON_ADD_RETURN" };
                }
                else
                {
                    allowedButtonsForLabelScreen = new string[1] { "MSG_BUTTON_DONE" };
                }

                goto Step_ScanGS1Label;
                
            }
            else if (msg.Name.EndsWith(".BackRequested"))
            {
                goto Step_SelectFilter;
            }
            else if (msg.Name.EndsWith(".ButtonClicked"))
            {
                buttonKey = ExtractParameter<string>(msg.Parameters, "buttonKey");

                if(buttonKey == "MSG_BUTTON_NO_LABEL")
                {
                    //No label
                    goto Step_GetProducts;
                }
                if(buttonKey == "MSG_BUTTON_ADD_RETURN")
                {
                    //Add return
                    goto Step_ScanDestination;
                }
                if (buttonKey == "MSG_BUTTON_DONE")
                {
                    //Done scanning
                    goto Step_GetProductFromBarCode;
                }
            }

            initialErrorKey = "MSG_ERROR_OPERATION_NOT_SUPPORTED";
            goto Step_ScanGS1Label;


         

            #endregion

        Step_GetProductFromBarCode:

            #region GET PRODUCT FROM BARCODE

            string[] dataFields = barcodeContainerObject.FreeBarcode.Split( ';' );
            try
            {
                if ( dataFields.Length == 5 )
                {
                    // Single pack
                    teoxLotNr = dataFields[ 1 ];
                    teoxFirstSeqNr = Convert.ToInt32( dataFields[ 2 ] );
                    teoxLastSeqNr = teoxFirstSeqNr;
                    teoxQuantity = 1;
                }
                else if ( dataFields.Length == 9 )
                {
                    // Full box
                    teoxLotNr = dataFields[ 2 ];
                    teoxFirstSeqNr = Convert.ToInt32( dataFields[ 4 ] );
                    teoxLastSeqNr = Convert.ToInt32( dataFields[ 5 ] );
                    teoxQuantity = Convert.ToInt32( dataFields[ 6 ] );
                }

                if ( teoxLotNr != null )
                {
                    DataTable tbl = new DataTable();
                    m_sboProviderService.InvokeMethodWithDbConnection<object>( false, false, null, null, delegate( PmxDbConnection conn, object[] parameters )
                    {
                        DbCommand dbCmd = conn.CreateCommand( TeoxaneBuildQuery.GetSerialNumberData( teoxLotNr, teoxFirstSeqNr, teoxLastSeqNr, false ) );
                        using ( DbDataReader rdr = dbCmd.ExecuteReader() )
                        {
                            tbl.Load( rdr );
                        }
                        return null;
                    } );

                    if ( tbl.Rows.Count < 1 )
                    {
                        initialErrorKey = "MSG_ERROR_UNKNOWN_SERIAL_NUMBERS";
                        goto Step_StartScanLabel;
                    }

                    // Multiple rows from PMX_TEOX_SERNUM not supported, but this should never happen anyway:
                    // - Shipped single sernum will be returned as single sernum
                    // - Shipped pack can be returned as pack or single
                    // - Shipped multiple single sernum's can not be returned as pack

                    int smallestRangeSize = Int32.MaxValue;
                    int? teoxSalesDeliveryDocEntry = null;
                    foreach ( DataRow row in tbl.Rows )
                    {
                        int currentFirstSeqNr = Convert.ToInt32( row[ PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ] );
                        int currentLastSeqNr = Convert.ToInt32( row[ PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME ] );
                        if ( currentFirstSeqNr >= teoxFirstSeqNr 
                            && currentLastSeqNr <= teoxLastSeqNr
                            && Tester.IsNullOrEmpty( row[ PmxLinkSalesDeliveryNoteLinePickListLineTable.Columns.SalesDeliveryNoteLineDocEntry.NAME ] ) )
                        {
                            // All found ranges equal or smaller than scanned range must be on a sales delivery, otherwise they can not be returned
                            initialErrorKey = "MSG_ERROR_SERIAL_NUMBERS_NOT_ON_SALES_DELIVERY";
                            goto Step_StartScanLabel;
                        }
                        if ( currentLastSeqNr - currentFirstSeqNr < smallestRangeSize
                            && currentFirstSeqNr <= teoxFirstSeqNr
                            && currentLastSeqNr >= teoxLastSeqNr )
                        {
                            // The smallest range fully containing scanned range is used as source data
                            smallestRangeSize = currentLastSeqNr - currentFirstSeqNr;

                            itemCode = Convert.ToString( row[ PmxPickListLineTable.Columns.ItemCode.NAME ] );
                            teoxPickListDocEntry = Convert.ToInt32( row[ PmxTeoxaneSerialNumbersTable.Columns.PickListDocEntry.NAME ] );
                            teoxPickListLineNum = Convert.ToInt32( row[ PmxTeoxaneSerialNumbersTable.Columns.PickListLineNum.NAME ] );
                            teoxBatchNum = Convert.ToString( row[ PmxItemTransactionalInfoTable.Columns.BatchNumber1.NAME ] );
                            teoxBatchNum2 = Convert.ToString( row[ PmxItemTransactionalInfoTable.Columns.BatchNumber2.NAME ] );
                            teoxSalesDeliveryDocEntry = SystemInt32Convertor.ObjectToIntOrNull( row[ PmxLinkSalesDeliveryNoteLinePickListLineTable.Columns.SalesDeliveryNoteLineDocEntry.NAME ] );
                        }
                    }

                    if ( !teoxSalesDeliveryDocEntry.HasValue )
                    {
                        initialErrorKey = "MSG_ERROR_SERIAL_NUMBERS_NOT_ON_SALES_DELIVERY";
                        goto Step_StartScanLabel;
                    }
                    if ( salesDeliveryDocEntry.HasValue && salesDeliveryDocEntry != teoxSalesDeliveryDocEntry )
                    {
                        initialErrorKey = "MSG_ERROR_SERIAL_NUMBERS_FROM_OTHER_SALES_DELIVERY";
                        goto Step_StartScanLabel;
                    }
                    salesDeliveryDocEntry = teoxSalesDeliveryDocEntry;  // Auto-select sales delivery
                    foreach ( ActionObject action in teoxSerialNumbersList )
                    {
                        if ( !action.RangeIsFree( teoxFirstSeqNr, teoxLastSeqNr ) )
                        {
                            initialErrorKey = ( teoxFirstSeqNr == teoxLastSeqNr )
                                ? "MSG_ERROR_SERIAL_NUMBER_ALREADY_SCANNED" : "MSG_ERROR_SERIAL_NUMBERS_OF_BOX_ALREADY_SCANNED";
                            goto Step_StartScanLabel;
                        }
                    }

                    goto Step_GetItemInfo;
                }
            }
            catch ( Exception ex )
            {
                initialErrorKey = ex.Message;
                goto Step_StartScanLabel;
            }
            
            if ( this.IsProductOnBarcode( barcodeContainerObject ) )
            {
                gtin = BarcodeHelper.GetProductIdentification(barcodeContainerObject);
                //check if item exists
                query = BuildQuery.GetProductForSalesDeliveryNoteOrderByName(gtin, null, null, m_sboProviderService.GetDbTool());
                dsItems = m_sboProviderService.RunView(false, null, null, query);

                if (this.GetNumberOfRows(dsItems) == 0)
                {
                    initialErrorKey = "MSG_ERROR_INCORRECT_PRODUCT";
                    goto Step_StartScanLabel;
                }

            }
            else
            {
                gtin = null;
            }

            goto Step_GetProducts;

            #endregion

        Step_GetProducts:

            #region GET PRODUCTS

            query = BuildQuery.GetProductForSalesDeliveryNoteOrderByName(gtin, sscc, salesDeliveryDocEntry, m_sboProviderService.GetDbTool());
            dsItems = m_sboProviderService.RunView(false, null, null, query);

            if (this.GetNumberOfRows(dsItems) == 0)
            {
                initialErrorKey = "MSG_ERROR_PRODUCT_NOT_NEEDED_BY_SALES_DELIVERY_NOTE";
                goto Step_StartScanLabel;
            }

            if (this.GetNumberOfRows(dsItems) == 1 &&
                !goBack)
            {
                //auto select it
                itemCode = dsItems.Tables[0].Rows[0][ProductDataSet.Columns.PRODUCT_CODE.Name] as string;
                itemDescription = dsItems.Tables[0].Rows[0][ProductDataSet.Columns.PRODUCT_DESCRIPTION.Name] as string;
                goto Step_GetItemInfo;
            }

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreen),
                DefaultCultureInfo.Get(), BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", "MSG_TITLE_SELECT_PRODUCT",
                    "ProductDS", dsItems
                    ));
            msg = WaitForMessage();

            //clear data
            initialErrorKey = null;
            goBack = false;

            if (msg.Name.EndsWith(".ProductSelected"))
            {
                itemCode = ExtractParameter<string>(msg.Parameters, "itemCode");
                //get itemdescription
                DataRow[] rows = dsItems.Tables[0].Select(string.Format("{0} = '{1}'", ProductDataSet.Columns.PRODUCT_CODE.Name, itemCode));
                if (rows.Length > 0)
                {
                    //copy the data
                    itemDescription = rows[0][ProductDataSet.Columns.PRODUCT_DESCRIPTION.Name] as string;
                }

                goto Step_GetItemInfo;
            }

            //Go back
            goto Step_StartScanLabel;

            #endregion

        Step_GetItemInfo:

            #region GET ITEM INFO

            itemInfo = m_sboProviderService.InvokeMethodWithDbConnection<PmxItemInfo>(false, false, null, null, new DelegateWithPmxDbConnection<PmxItemInfo>(ExecuteGetItemInfo), itemCode);

            uom = itemInfo.SalesUom;
            quantityPerUom = itemInfo.NumberOfItemsPerSalesUnit;
            innerUom = itemInfo.InventoryUom;

            if (salesDeliveryDocEntry.HasValue)
            {

                openQuantityList = m_sboProviderService.InvokeMethodWithDbConnection<Collection<OpenQuantityBySalesDeliveryLineNum>>(false, false, null, null,
                    delegate(PmxDbConnection conn, object[] parameters)
                    {
                        Collection<OpenQuantityBySalesDeliveryLineNum> list = new Collection<OpenQuantityBySalesDeliveryLineNum>(); ;
                        PmxCommandProviderInternal viewProv = new PmxCommandProviderInternal(conn);
                        string qry = SboCommands.GetQueryGetOpenQtyFromSalesDelivery(salesDeliveryDocEntry.Value, itemCode, sscc, conn.DbCultureInfo);
                        using (ISboRecordset rs = viewProv.RunQuery(qry))
                        {
                            while (!rs.EoF)
                            {
                                OpenQuantityBySalesDeliveryLineNum openQty = new OpenQuantityBySalesDeliveryLineNum();
                                openQty.OpenQuantity = Convert.ToDouble(rs[SboDocumentLineColumns.OpenQty.NAME]);
                                openQty.LineNum = Convert.ToInt32(rs[SboDocumentLineColumns.LineNum.NAME]);

                                list.Add(openQty);

                                rs.MoveNext();
                            }
                        }
                        return list;

                    });
            }
            else
            {
                openQuantityList = null;
            }

            goto Step_SelectBatchNumber;


            #endregion

        Step_SelectBatchNumber:

            #region SELECT BATCH NUMBER

            batchNumber = null;
            batchNumber2 = null;
            bestBeforeDate = null;
            itriKey = null;

            string tempBatchNumber = null;
            if ( teoxLotNr != null )
            {
                tempBatchNumber = teoxBatchNum;
            }
            else if ( barcodeContainerObject != null && barcodeContainerObject.GS1DataObject != null
                && !string.IsNullOrEmpty( barcodeContainerObject.GS1DataObject.LotNumber ) )
            {
                tempBatchNumber = barcodeContainerObject.GS1DataObject.LotNumber.ToUpper();
            }

            DateTime? tempBBD = null;
            if ( barcodeContainerObject != null && barcodeContainerObject.GS1DataObject != null )
            {
                tempBBD = barcodeContainerObject.GS1DataObject.BestBeforeDate;
            }

            if (salesDeliveryDocEntry.HasValue)
            {
                #region BATCH BBD FOR DELIVERY


                query = BuildQuery.GetBatchNumbersForSalesRetour(salesDeliveryDocEntry.Value, itemCode, tempBatchNumber, tempBBD, sscc);
                dsBatchNumbers = m_sboProviderService.RunView(false, null, null, query);

                dsBatchNumbers = RemoveReturnedBatches(salesReturnItemInfoList, itemCode, dsBatchNumbers);

                if (this.GetNumberOfRows(dsBatchNumbers) == 0)
                {
                    initialErrorKey = "MSG_ERROR_NO_BATCH_NUMBERS_FOUND";
                    goto Step_StartScanLabel;
                }

                if (this.GetNumberOfRows(dsBatchNumbers) == 1 &&
                    !goBack)
                {

                    batchNumber = dsBatchNumbers.Tables[0].Rows[0][BatchNumberDataSet.Columns.BATCH_NUMBER.Name] as string;
                    batchNumber2 = dsBatchNumbers.Tables[0].Rows[0][BatchNumberDataSet.Columns.BATCH_NUMBER2.Name] as string;
                    object bbd = dsBatchNumbers.Tables[0].Rows[0][BatchNumberDataSet.Columns.BEST_BEFORE_DATE.Name];
                    if (bbd is DateTime)
                    {
                        bestBeforeDate = Convert.ToDateTime(bbd);
                    }
                    else
                    {
                        bestBeforeDate = null;
                    }

                    object itri = dsBatchNumbers.Tables[0].Rows[0][BatchNumberDataSet.Columns.ITEMTRANSACTIONALINFOKEY.Name];
                    if (itri is int)
                    {
                        itriKey = Convert.ToInt32(itri);
                    }
                    else
                    {
                        itriKey = null;
                    }

                    batchQuantity = Convert.ToDouble(dsBatchNumbers.Tables[0].Rows[0][BatchNumberDataSet.Columns.QUANTITY.Name]);

                    goto Step_SetQuantities;

                }

                //more than 1 batchnumber, so select it
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectBatchNumberScreen),
                DefaultCultureInfo.Get(), BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", "MSG_TITLE_SELECT_BATCH_NUMBER",
                    "BatchNumberDS", dsBatchNumbers
                    ));
                msg = WaitForMessage();

                //clear data
                initialErrorKey = null;
                goBack = false;

                if (msg.Name.EndsWith(".BatchNumberSelected"))
                {
                    batchNumber = ExtractParameter<string>(msg.Parameters, "batchNumber");
                    batchNumber2 = ExtractParameter<string>(msg.Parameters, "batchNumber2");
                    bestBeforeDate = ExtractParameter<DateTime?>(msg.Parameters, "bestBeforeDate");
                    batchQuantity = ExtractParameter<double>(msg.Parameters, "quantity");
                    itriKey = ExtractParameter<int?>(msg.Parameters, "itemTransactionalInfoKey");

                    goto Step_SetQuantities;
                }

                //Go back
                goBack = true;
                goto Step_GetProducts;


                #endregion
            }
            else
            {
                #region BATCH NO DELIVERY

                if (itemInfo.HasBatchnumber)
                {
                    if (string.IsNullOrEmpty(tempBatchNumber))
                    {

                        //no batchnumber provided, so ask for batchnumber
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Reception.IIdentifyBatchNumberScreen),
                         DefaultCultureInfo.Get(), BuildParamCollection(
                             "InitialErrorKey", initialErrorKey,
                             "TitleKey", "MSG_TITLE_IDENTIFY_BATCH_NUMBER",
                             "ItemDescription", itemDescription,
                             "ItemCode", itemInfo.ItemCode,
                             "ItemBarcode", itemInfo.CodeBars,
                             "CurrentSSCC", sscc,
                             "Uom", string.IsNullOrEmpty(itemInfo.SalesUom) ? itemInfo.InventoryUom : itemInfo.SalesUom
                            ));
                        msg = WaitForMessage();

                        // reset error
                        initialErrorKey = null;
                        goBack = false;

                        if (msg.Name.EndsWith(".BatchNumberEntered"))
                        {
                            batchNumber = ExtractParameter<string>(msg.Parameters, "batchNumber");
                            goto Step_SecondBatchNumber;
                        }

                        //Go back
                        goBack = true;
                        goto Step_GetProducts;

                    }
                    else
                    {
                        batchNumber = tempBatchNumber;
                        goto Step_SecondBatchNumber;
                    }




                }
                else
                {
                    if (goBack)
                    {
                        goto Step_GetProducts;
                    }
                    //no batchnumber, so we try to get second batch number
                    goto Step_SecondBatchNumber;
                }
                #endregion
            }

            #endregion

        Step_SecondBatchNumber:

            #region ENTER SECOND BATCH NUMBER

            batchNumber2 = null;

            if (itemInfo.HasSecondBatchNumber)
            {
                //no batchnumber provided, so ask for batchnumber
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Reception.IIdentifyBatchNumberScreen),
                 DefaultCultureInfo.Get(), BuildParamCollection(
                     "InitialErrorKey", initialErrorKey,
                     "TitleKey", "MSG_TITLE_IDENTIFY_BATCH_NUMBER2",
                     "ItemDescription", itemDescription,
                     "ItemCode", itemInfo.ItemCode,
                     "ItemBarcode", itemInfo.CodeBars,
                     "CurrentSSCC", sscc,
                     "Uom", string.IsNullOrEmpty(itemInfo.SalesUom) ? itemInfo.InventoryUom : itemInfo.SalesUom
                    ));
                msg = WaitForMessage();

                // reset error
                initialErrorKey = null;

                if (msg.Name.EndsWith(".BatchNumberEntered"))
                {
                    batchNumber2 = ExtractParameter<string>(msg.Parameters, "batchNumber");
                    goto Step_BestBeforeDate;
                }
                else if (msg.Name.EndsWith(".BackRequested"))
                {
                    goBack = true;
                    goto Step_SelectBatchNumber;
                }



            }
            else
            {
                if (goBack)
                {
                    goto Step_SelectBatchNumber;
                }
                //no second batchnumber, so we try to get best before date
                goto Step_BestBeforeDate;
            }

            #endregion

        Step_BestBeforeDate:

            #region PROVIDE BEST BEFORE DATE
            if (itemInfo.HasBestBeforeDate)
            {
                if (barcodeContainerObject != null &&
                    barcodeContainerObject.GS1DataObject != null &&
                    barcodeContainerObject.GS1DataObject.BestBeforeDate.HasValue)
                {
                    if (goBack)
                    {
                        if (itemInfo.HasSecondBatchNumber)
                        {
                            goto Step_SecondBatchNumber;
                        }
                        else if (itemInfo.HasBatchnumber)
                        {
                            goto Step_SelectBatchNumber;
                        }

                        goto Step_GetProducts;


                    }
                    bestBeforeDate = barcodeContainerObject.GS1DataObject.BestBeforeDate.Value;

                    goto Step_SetQuantities;

                }
                else
                {
                    goBack = false;
                    //ask for best before date
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.Reception.IEnterBestBeforeDateScreen),
                     DefaultCultureInfo.Get(), BuildParamCollection(
                         "InitialErrorKey", initialErrorKey,
                         "TitleKey", "MSG_TITLE_ENTER_BEST_BEFORE_DATE",
                         "ItemDescription", itemDescription,
                         "ItemCode", itemInfo.ItemCode,
                         "ItemBarcode", itemInfo.CodeBars,
                         "CurrentSSCC", sscc,
                         "Uom", string.IsNullOrEmpty(itemInfo.SalesUom) ? itemInfo.InventoryUom : itemInfo.SalesUom,
                         "BatchNumber", batchNumber
                         ));
                    msg = WaitForMessage();

                    // reset error
                    initialErrorKey = null;

                    if (msg.Name.EndsWith(".BestBeforDateEntered"))
                    {
                        bestBeforeDate = ExtractParameter<DateTime?>(msg.Parameters, "bestBeforeDate");

                        goto Step_SetQuantities;

                    }

                    goBack = true;
                    if (itemInfo.HasSecondBatchNumber)
                    {
                        goto Step_SecondBatchNumber;
                    }
                    else if (itemInfo.HasBatchnumber)
                    {
                        goto Step_SelectBatchNumber;
                    }
                    goto Step_GetProducts;

                }
            }
            else
            {
                if (goBack)
                {
                    if (itemInfo.HasSecondBatchNumber)
                    {
                        goto Step_SecondBatchNumber;
                    }
                    else if (itemInfo.HasBatchnumber)
                    {
                        goto Step_SelectBatchNumber;
                    }
                    goto Step_GetProducts;
                }
                //no best before date
                goto Step_SetQuantities;
            }

            #endregion

        Step_SetQuantities:

            #region SET QUANTITIES

            double qtyToReduce = 0;

            if (salesDeliveryDocEntry.HasValue)
            {
                query = BuildQuery.GetQueryGetReturnedQtyFromSalesDelivery(salesDeliveryDocEntry.Value, itemCode, batchNumber, bestBeforeDate);
                ds = m_sboProviderService.RunView(false, null, null, query);

                if (this.GetNumberOfRows(ds) > 0)
                {
                    qtyToReduce = Convert.ToDouble(ds.Tables[0].Rows[0][PmxInventoryDetailTable.Columns.Quantity.NAME]);
                }


                batchQuantity -= qtyToReduce;

                maximumQuantity = batchQuantity;

                if (maximumQuantity <= 0)
                {
                    initialErrorKey = "MSG_ERROR_CANNOT_RETURN_ITEM";
                    goto Step_StartScanLabel;
                }
            }
            else
            {
                maximumQuantity = -1;
            }

            
            goto Step_EnterQuantity;

            #endregion

        Step_EnterQuantity:

            #region ENTER QUANTITY

            forceSerial = false;
            useRange = false;
            listOfSerialNumbers = null;

            if ( teoxQuantity > 0 )
            {
                quantity = teoxQuantity;

                goto Step_SelectReason;
            }

            if (itemInfo.HasSerialNumber)
            {
                forceSerial = true;
                //show screen to ask user what to do.
                buttonKeys = new string[2] { "MSG_BUTTON_SERIAL_NUMBER", "MSG_BUTTON_SERIAL_NUMBER_RANGE" };

                toolBarButtons = ToolBarButtons.Back;

                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectButtonScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                        "InitialErrorKey", initialErrorKey,
                        "TitleKey", "MSG_TITLE_ENTER_QUANTITY_SERIAL_NUMBER_TASK",
                        "AllowedButtons", buttonKeys,
                        "EnabledToolBarButtons", toolBarButtons
                        ));
                msg = WaitForMessage();
                initialErrorKey = null;
                if (msg.Name.EndsWith(".ButtonClicked"))
                {
                    filterButtonKey = ExtractParameter<string>(msg.Parameters, "buttonKey");

                    useRange = filterButtonKey == "MSG_BUTTON_SERIAL_NUMBER_RANGE";
                }
                else
                {
                    //Go back
                    goBack = true;
                    goto Step_BestBeforeDate;
                }

            }



            WorkflowInstanceBase subEnterQuantityFlow = this.Create("EnterQuantityScript");
            m_subFlowBackRequested = false;
            subEnterQuantityFlow.BindParameters(
                this.DefaultCultureInfo,
                this.PmxOseCompany,
                new ReadOnlyBinder<string>("TitleKey", "MSG_TITLE_ENTER_QUANTITY"),
                new ReadOnlyBinder<PmxItemInfo>("ItemInfo", itemInfo),
                new ReadOnlyBinder<int?>("ItriKey", itriKey),
                new ReadOnlyBinder<string>("BatchNumber", batchNumber),
                new ReadOnlyBinder<DateTime?>("BestBeforeDate", bestBeforeDate),
                new ReadOnlyBinder<string>("SSCC", sscc),

                new ReadOnlyBinder<bool>("AllowZero", false),
                new ReadOnlyBinder<double>("InitialQuantity", 1),
                new ReadOnlyBinder<double>("MaximumQuantity", maximumQuantity),
                new ReadOnlyBinder<string>("UOM", uom),
                new ReadOnlyBinder<int>("NumberOfDigitsAfterDecimal", (itemInfo == null) ? 0 : itemInfo.UomDecimals),
                new ReadOnlyBinder<double>("InnerQuantity", quantityPerUom),
                new ReadOnlyBinder<string>("InnerUOM", innerUom),

                new ReadOnlyBinder<string>("DocNum", docNum),
                new ReadOnlyBinder<double>("StillToQuantity", maximumQuantity),
                new ReadOnlyBinder<string>("MessageKeyStillQuantity", "MSG_LBL_MAXIMUM_QUANTITY"),

                new ReadOnlyBinder<bool>("ForceInputSerialNumbers", forceSerial),
                new ReadOnlyBinder<string>("ObjectType", ((int)SboObjectType.SalesReturn).ToString()),
                new ReadOnlyBinder<int?>("DocEntry", salesDeliveryDocEntry),
                new ReadOnlyBinder<bool>("UseSerialNumberRange", useRange),
                new ReadOnlyBinder<string>("LocationCode", deviceID),


                new ReadWriteBinder<bool>("BackRequested", this, "m_subFlowBackRequested"),
                new ReadWriteBinder<double>("QuantityEntered", this, "m_subFlowQuantity"),
                new ReadWriteBinder<Collection<string>>("ListOfSerialNumbers", this, "m_subFlowListOfSerialNumbers")
            );

            

            base.StartSubFlowAndWait(subEnterQuantityFlow, true);


            if (m_subFlowBackRequested)
            {
                goBack = true;
                goto Step_BestBeforeDate;
            }

            listOfSerialNumbers = m_subFlowListOfSerialNumbers;
            quantity = m_subFlowQuantity;

            goto Step_SelectReason;

            #endregion

        Step_SelectReason:

            #region SELECT REASON

            if (m_sboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null,
                    delegate(PmxDbConnection conn, object[] parameters)
                    {
                        PmxItemAllConnectionsProvider prov = new PmxItemAllConnectionsProvider(conn);
                        return prov.DoesItemNeedReasonForSalesReturn(itemCode);
                    }))
            {
                //select reason
                query = BuildQuery.GetSalesRetourReasons();
                dsReasons = m_sboProviderService.RunView(false, null, null, query);

                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectReasonScreen),
                     this.DefaultCultureInfo, BuildParamCollection(
                     "InitialErrorKey", initialErrorKey,
                     "TitleKey", "MSG_TITLE_SELECT_REASON",
                     "MessageKey", "MSG_SELECT_SALES_RETOUR_REASON",
                     "ReasonDS", dsReasons));

                msg = WaitForMessage();
                initialErrorKey = null;

                if (msg.Name.EndsWith(".ReasonSelected"))
                {
                    reasonCode = ExtractParameter<string>(msg.Parameters, "reasonCode");
                    goto Step_AddToList;
                }

                //go back
                goto Step_EnterQuantity;
            }

            goto Step_AddToList;

            #endregion

        Step_AddToList:

            #region ADD TO LIST

            if (salesReturnItemInfoList == null)
            {
                salesReturnItemInfoList = new Collection<SalesReturnItemInfo>();
            }

            //first check if there is openqty
            if ( openQuantityList != null )
            {
                double qtyStillToReturn = quantity;
                foreach ( OpenQuantityBySalesDeliveryLineNum openQty in openQuantityList )
                {
                    if ( openQty.OpenQuantity > 0 )
                    {
                        SalesReturnItemInfo retInfo = new SalesReturnItemInfo();
                        retInfo.Quantity = Math.Min( qtyStillToReturn, openQty.OpenQuantity );
                        //reduce the still to return quantity)
                        qtyStillToReturn -= retInfo.Quantity;
                        if (!string.IsNullOrEmpty(reasonCode))
                        {
                            retInfo.ReasonInfo = new ReasonInfo();
                            retInfo.ReasonInfo.ReasonCode = reasonCode;
                        }
                        retInfo.BatchNumber = batchNumber;
                        retInfo.BatchNumber2 = batchNumber2;
                        retInfo.BestBeforeDate = bestBeforeDate;
                        retInfo.ItemCode = itemCode;
                        //Take the description of SAP
                        retInfo.ItemDescription = itemInfo.Description;
                        retInfo.Uom = uom;
                        retInfo.QuantityPerUom = quantityPerUom;

                        retInfo.LinkType = Produmex.Foundation.Data.Sbo.BusinessObjects.SboLinkDocumentType.SalesDeliveryNote;
                        retInfo.LinkDocEntry = salesDeliveryDocEntry;
                        retInfo.LinkLineNum = openQty.LineNum;

                        retInfo.ListOfSerialNumbers = listOfSerialNumbers;

                        salesReturnItemInfoList.Add( retInfo );

                        ActionObject teoxSerialNumber = new ActionObject();
                        teoxSerialNumber.SalesReturnItemInfo = retInfo;
                        if ( teoxQuantity != -1 )
                        {
                            teoxSerialNumber.PickListDocEntry = teoxPickListDocEntry;
                            teoxSerialNumber.PickListLineNum = teoxPickListLineNum;
                            teoxSerialNumber.AddRange( teoxFirstSeqNr, teoxLastSeqNr, teoxQuantity );
                        }
                        teoxSerialNumbersList.Add( teoxSerialNumber );

                        if ( qtyStillToReturn <= 0 )
                        {
                            break;
                        }
                    }
                }//foreach
            }
            else
            {
                SalesReturnItemInfo retInfo = new SalesReturnItemInfo();
                retInfo.Quantity = quantity;

                if (!string.IsNullOrEmpty(reasonCode))
                {
                    retInfo.ReasonInfo = new ReasonInfo();
                    retInfo.ReasonInfo.ReasonCode = reasonCode;
                }

                retInfo.BatchNumber = batchNumber;
                retInfo.BatchNumber2 = batchNumber2;
                retInfo.BestBeforeDate = bestBeforeDate;
                retInfo.ItemCode = itemCode;
                retInfo.ItemDescription = itemDescription;
                retInfo.Uom = uom;
                retInfo.QuantityPerUom = quantityPerUom;
                retInfo.ListOfSerialNumbers = listOfSerialNumbers;
                salesReturnItemInfoList.Add( retInfo );

                ActionObject teoxSerialNumber = new ActionObject();
                teoxSerialNumber.SalesReturnItemInfo = retInfo;
                if ( teoxQuantity != -1 )
                {
                    teoxSerialNumber.PickListDocEntry = teoxPickListDocEntry;
                    teoxSerialNumber.PickListLineNum = teoxPickListLineNum;
                    teoxSerialNumber.AddRange( teoxFirstSeqNr, teoxLastSeqNr, teoxQuantity );
                }
                teoxSerialNumbersList.Add( teoxSerialNumber );
            }

            #endregion

        Step_AskToAddMore:

            #region ASK TO ADD MORE

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                   DefaultCultureInfo.Get(), BuildParamCollection(
                       "TitleKey", "MSG_TITLE_ADD_MORE_PRODUCTS",
                       "MessageKey", "MSG_QUESTION_ADD_MORE_RETURNS"));
            msg = WaitForMessage();

            if (msg.Name.ToLower().EndsWith(".yes"))
            {
                //add next item
                goto Step_StartScanLabel;
            }

            goto Step_ScanDestination;

            #endregion
            
        Step_ScanDestination:

            #region SCAN DESTINATION

            newSSCC = false;
            emptyLocations = false;
            dstLuid = null;
            targetLogCarrierCode = null;
            destinationPmxWhsCode = null;
			blockMoveWhenNotEmpty = false;
            dstLocationCode = null;

            #endregion

        Step_ScanDestinationAgain:

            #region SCAN DESTINATION AGAIN

            string[] buttons = null;

            if (newSSCC)
            {
                buttons = new string[] { "MSG_BUTTON_SELECT_LOCATION", "MSG_BUTTON_SELECT_EMPTY_LOCATION" };
            }
            else
            {
                buttons = new string[] { "MSG_BUTTON_SELECT_LOCATION", "MSG_BUTTON_SELECT_EMPTY_LOCATION", "MSG_BUTTON_NEW_SSCC" };
            }

            session.ShowScreen( typeof( IScanStringValueScreen ),
                this.DefaultCultureInfo, BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", newSSCC ? "MSG_TITLE_SCAN_DESTINATION_LOCATION" : "MSG_TITLE_DST_LOCATION_OR_SSCC",
                    "ForceDataEntry", true,
                    "AllowedButtonKeys", buttons,
                    "AllowToGoBack", true
                    ));

            initialErrorKey = null;
            msg = WaitForMessage();

            if ( msg.Name.EndsWith(".StringScanned") )
            {
                BarcodeContainer barcode = new BarcodeContainer();
                ProdumexError pmxError = null;
                try
                {
                    pmxError = BarcodeHelper.ParseElementString( ExtractParameter<string>( msg.Parameters, "stringValue" ), true, barcode );
                }
                catch
                {
                    // Do nothing
                }
                if ( !String.IsNullOrEmpty( barcode.GS1DataObject.Sscc ) )
                {
                    query = BuildQuery.GetSSCCInfo( barcode.GS1DataObject.Sscc );
                    ds = m_sboProviderService.RunView( true, null, null, query );
                    if ( this.GetNumberOfRows( ds ) > 0 )
                    {
                        newSSCC = false;

                        dstLuid = Convert.ToInt32( ds.Tables[ 0 ].Rows[ 0 ][ SSCCDataSet.Columns.LUID.Name ] );
                        dstLocationCode = Convert.ToString(ds.Tables[ 0 ].Rows[ 0 ][ SSCCDataSet.Columns.STORAGE_LOCATION_CODE.Name ]);
                        destinationPmxWhsCode = Convert.ToString(ds.Tables[0].Rows[0][SSCCDataSet.Columns.PMX_WAREHOUSE_CODE.Name]);
                    }
                    if ( String.IsNullOrEmpty( dstLocationCode ) )
                    {
                        initialErrorKey = "MSG_ERROR_INCORRECT_SSCC_TO_RETURN";
                        goto Step_ScanDestination;
                    }

                    goto Step_GetDstLocation;
                }
                else
                {
                    dstLocationCode = ExtractParameter<string>( msg.Parameters, "stringValue" );
                    goto Step_GetDstLocation;
                }
            }
            if ( msg.Name.EndsWith(".ButtonClicked" ))
            {
                buttonKey = ExtractParameter<string>( msg.Parameters, "buttonKey" );
                if ( buttonKey == "MSG_BUTTON_SELECT_LOCATION" )
                {
                    goto Step_GetDstLocation;
                }
                if ( buttonKey == "MSG_BUTTON_SELECT_EMPTY_LOCATION" )
                {
                    emptyLocations = true;
                    goto Step_GetDstLocation;
                }
                if ( buttonKey == "MSG_BUTTON_NEW_SSCC" )
                {
                    #region Step_IdentifyLogisticCarrier

                    query = BuildQuery.GetLogisticCarriersOrderByName( true, true, m_sboProviderService.GetDbTool() );
                    DataSet dsLogisticCarriers = m_sboProviderService.RunView( false, null, null, query );
                    if ( this.GetNumberOfRows( dsLogisticCarriers ) > 0 )
                    {
                        session.ShowScreen( typeof( Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreen ),
                            this.DefaultCultureInfo, BuildParamCollection(
                            "InitialErrorKey", initialErrorKey,
                            "TitleKey", "MSG_TITLE_SELECT_LOGISTIC_CARRIER",
                            "ProductDS", dsLogisticCarriers,
                            "AllowNoSelection", false,
                            "ButtonKey", "MSG_BUTTON_NO_LOGISTIC_CARRIER" ) );
                        msg = WaitForMessage();
                        initialErrorKey = null;

                        if ( msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreenReply.ProductSelected" )
                        {
                            targetLogCarrierCode = ExtractParameter<string>( msg.Parameters, "itemCode" );
                        }
                        else if ( msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreenReply.ButtonClicked" )
                        {
                            targetLogCarrierCode = "";  // Remember we chose MSG_BUTTON_NO_LOGISTIC_CARRIER
                        }
                        else if ( msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreenReply.BackRequested" )
                        {
                            goto Step_ScanDestination;
                        }
                        else
                        {
                            initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                            goto Step_ScanDestination;
                        }
                    }

                    #endregion

                    newSSCC = true;
                    goto Step_ScanDestinationAgain;
                }
            }

            //go back
            goto Step_StartScanLabel;
            

            //goto Step_ScanDestinationAgain;

            #endregion

        Step_GetDstLocation:

            #region GET DESTINATION LOCATION

            //get list of items/quantities
            canItemBeStoredInfoPerItemCode = new Dictionary<string, PmxStorageLocationCanItemBeStoredInfo>();
            foreach ( SalesReturnItemInfo info in salesReturnItemInfoList )
            {
                PmxStorageLocationCanItemBeStoredInfo entry;
                if ( canItemBeStoredInfoPerItemCode.TryGetValue( info.ItemCode, out entry ) )
                {
                    entry.Quantity += info.Quantity;
                }
                else
                {
                    entry = new PmxStorageLocationCanItemBeStoredInfo( info.Quantity, null, null );
                    canItemBeStoredInfoPerItemCode.Add( info.ItemCode, entry );
                }
                if ( !String.IsNullOrEmpty( info.BatchNumber ) )
                {
                    if ( entry.BatchNums == null ) entry.BatchNums = new List<string>();
                    entry.BatchNums.Add( info.BatchNumber );
                }
            }

            itemCodes = new Collection<string>();
            foreach ( string temp in canItemBeStoredInfoPerItemCode.Keys )
            {
                itemCodes.Add(temp);
            }

            query = BuildQuery.GetZoneTypesForItems(itemCodes);
            ds = m_sboProviderService.RunView(false, null, null, query );
            List<string> zoneTypeCodes = null;
            if (this.GetNumberOfRows(ds) > 0)
            {
                zoneTypeCodes = new List<string>(this.GetNumberOfRows(ds));
                foreach (DataRow row in ds.Tables[0].Rows)
                {
                    zoneTypeCodes.Add(Convert.ToString(row[PmxItemZoneTypeTable.Columns.ZoneTypeCode.NAME]));
                }
            }

            query = BuildQuery.GetStorageLocationForZoneTypes( dstLocationCode, whsInfo.PmxWarehouseCode, zoneTypeCodes, emptyLocations );
            ds = m_sboProviderService.RunView( true, null, null, query );
            if (this.GetNumberOfRows(ds) == 0)
            {
                initialErrorKey = "MSG_ERROR_NO_LOCATION_FOUND";
                goto Step_ScanDestination;
            }
            if (this.GetNumberOfRows(ds) == 1)
            {
                //set locationcode from DB to avoid problems with uppercase/lowercase
                dstLocationCode = Convert.ToString(ds.Tables[0].Rows[0][LocationDataSet.Columns.CODE.Name]);
                destinationPmxWhsCode = Convert.ToString(ds.Tables[0].Rows[0][LocationDataSet.Columns.WAREHOUSECODE.Name]);
                blockMoveWhenNotEmpty = Convert.ToString(ds.Tables[0].Rows[0][PmxOseStorageLocationGeneralTable.Columns.BlockMoveWhenNotEmpty.NAME]) == SboBooleanColumnValues.Y.VALUE;

                goto Step_VerifyDestination;
            }

            session.ShowScreen( typeof( Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreen ),
                this.DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_SELECT_DESTINATION_LOCATION",
                "LocationDS", ds ) );
            initialErrorKey = null;
            msg = WaitForMessage();

            if ( msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreenReply.LocationSelected" )
            {
                dstLocationCode = ExtractParameter<string>( msg.Parameters, "locationCode" );
                destinationPmxWhsCode = ExtractParameter<string>(msg.Parameters, "warehouseCode");
				DataRow[] rows = ds.Tables[0].Select(string.Format("{0} = '{1}'", LocationDataSet.Columns.CODE.Name, dstLocationCode));

                if (rows.Length > 0)
                {
                    blockMoveWhenNotEmpty = Convert.ToString(rows[0][PmxOseStorageLocationGeneralTable.Columns.BlockMoveWhenNotEmpty.NAME]) == SboBooleanColumnValues.Y.VALUE;
                }
                goto Step_VerifyDestination;
            }

            goto Step_ScanDestination;

            #endregion

        Step_VerifyDestination:

            #region VERIFY DESTINATION

            isLocked = m_sboProviderService.InvokeMethodWithDbConnection<bool>( false, true, null, null, delegate( PmxDbConnection conn, object[] parameters )
            {
                PmxOseStorageLocationProvider prov = new PmxOseStorageLocationProvider( conn );
                return prov.IsLocationLocked( dstLocationCode );
            } );

            if ( isLocked )
            {
                initialErrorKey = "MSG_ERROR_DESTINATION_LOCATION_LOCKED";
                goto Step_ScanDestination;
            }

            if (blockMoveWhenNotEmpty)
            {
                bool locationEmpty = m_sboProviderService.InvokeMethodWithDbConnection<bool>(false, true, null, null, delegate(PmxDbConnection conn, object[] parameters)
                {
                    PmxInventoryProvider prov = new PmxInventoryProvider(conn);
                    double? dummy;
                    double tempQty = prov.GetCurrentQuantityForLocation(dstLocationCode, out dummy);

                    return tempQty == 0;
                });

                if (!locationEmpty)
                {
                    initialErrorKey = "MSG_ERROR_ITEMS_CANNOT_BE_STORED_ON_LOCATION_NOT_EMPTY";
                    goto Step_ScanDestination;
                }
            }

            bool canBeStored = m_sboProviderService.InvokeMethodWithDbConnection<bool>( false, true, null, null, delegate( PmxDbConnection conn, object[] parameters )
            {
                PmxOseStorageLocationProvider prov = new PmxOseStorageLocationProvider( conn );
                return prov.CanItemsBeStoredOnStorageLocation( dstLocationCode, canItemBeStoredInfoPerItemCode );
            } );

            if ( !canBeStored )
            {
                initialErrorKey = "MSG_ERROR_ITEMS_CANNOT_BE_STORED_ON_LOCATION";
                goto Step_ScanDestination;
            }

            goto Step_AddReturn;

            #endregion

        Step_AddReturn:

            #region ADD RETURN

            salesReturnDocEntry = null;

            try
            {

                salesReturnDocEntry = m_sboProviderService.InvokeMethodWithDbConnection<int>(true, false, null, null,
                    delegate(PmxDbConnection conn, object[] parameters)
                    {
                        if (newSSCC)
                        {
                            PmxLogisticUnitIDProvider luidProv = new PmxLogisticUnitIDProvider(conn);
                            dstLuid = luidProv.GenerateNewLogisticUnit(null);
                        }
                        PmxSalesReturnProvider prov = new PmxSalesReturnProvider((SboConnection)conn);
                        int newSalesReturnDocEntry = prov.AddSalesReturn( cardCode, dstLocationCode, salesReturnItemInfoList, dstLuid, targetLogCarrierCode );

                        DbCommand dbCmd = conn.CreateCommand( TeoxaneBuildQuery.GetSalesReturnLinesBaseInfo( newSalesReturnDocEntry ) );
                        SortedList<string, int> baseInfoToLineNum = new SortedList<string, int>( salesReturnItemInfoList.Count );
                        using ( DbDataReader rdr = dbCmd.ExecuteReader() )
                        {
                            while ( rdr.Read() )
                            {
                                string key = String.Format( "{0}|{1}|{2}", rdr[ PmxSalesReturnLineTable.Columns.BaseType.NAME ], 
                                    rdr[ PmxSalesReturnLineTable.Columns.BaseEntry.NAME ], rdr[ PmxSalesReturnLineTable.Columns.BaseLine.NAME ] );
                                if ( !key.StartsWith( "-1|" ) )
                                {
                                    baseInfoToLineNum[ key ] = Convert.ToInt32( rdr[ PmxSalesReturnLineTable.Columns.LineNum.NAME ] );
                                }
                            }
                        }

                        foreach ( ActionObject action in teoxSerialNumbersList )
                        {
                            action.SalesReturnDocEntry = newSalesReturnDocEntry;
                            string key = String.Format( "{0}|{1}|{2}", (int)action.SalesReturnItemInfo.LinkType,
                                action.SalesReturnItemInfo.LinkDocEntry, action.SalesReturnItemInfo.LinkLineNum );
                            if ( !key.StartsWith( "-1|" ) )
                            {
                                action.SalesReturnLineNum = baseInfoToLineNum[ key ];
                            }
                            action.PreAction( conn );
                        }

                        return newSalesReturnDocEntry;
                    });

                try
                {
                    //just log, no actions required
                    m_sboProviderService.InvokeMethodWithDbConnection<bool>(true, false, null, null,
                    delegate(PmxDbConnection conn, object[] parameters)
                    {
                        PmxMoveProvider prov = new PmxMoveProvider((SboConnection)conn);
                        return prov.CreateMoveForLogisticUnitsWithOnlyLogicCarriersIfConfigured(dstLocationCode);
                    });
                }
                catch (Exception ex2)
                {
                    s_log.Error("", ex2);
                }


                numberOfReports = 0;

                // Print LABEL
                if (newSSCC)
                {

                    //print the label
                    bool proceed = false;
                    while (!proceed)
                    {
                        numberOfReports = 0;

                        try
                        {
                            //trigger event
                            numberOfReports = m_sboProviderService.InvokeMethodWithDbConnection<int>(false, true, null, null, delegate(PmxDbConnection conn, object[] parameters)
                            {
                                PmxPrintReportProvider printProv = new PmxPrintReportProvider(conn);
                                return printProv.PrintReportsForEvent(PmxPrintReportEventType.WarehouseCreatedLogisticUnit, dstLocationCode, session.DeviceInstanceId, dstLuid.Value, out askReprint);
                            });
                        }
                        catch (Produmex.Foundation.ProdumexException ex)
                        {
                            s_log.Error(ex.ToString());
                            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                                DefaultCultureInfo.Get(), BuildParamCollection(
                                    "InitialErrorKey", initialErrorKey,
                                    "MessageKey", KnownMessageKeys.MSG_ERROR_INTERNAL_ERROR,
                                    "MessageAddendum", ex.Message,
                                    "PmxError", ex.Error,
                                    "ShowButton", true));

                            msg = WaitForMessage();
                        }
                        catch (Exception ex)
                        {
                            s_log.Error(ex.ToString());
                            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                                DefaultCultureInfo.Get(), BuildParamCollection(
                                    "InitialErrorKey", initialErrorKey,
                                    "MessageKey", KnownMessageKeys.MSG_ERROR_INTERNAL_ERROR,
                                    "MessageAddendum", ex.Message,
                                    "PmxError", Produmex.Sbo.Logex.Data.ProdumexErrors.UnexpectedError(s_log, ex),
                                    "ShowButton", true));

                            msg = WaitForMessage();
                        }

                        if (numberOfReports > 0 && askReprint)
                        {
                            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                                DefaultCultureInfo.Get(), BuildParamCollection(
                                    "MessageKey", "MSG_QUESTION_REPRINT_LABEL",
                                    "TitleKey", "MSG_TITLE_REPRINT_LABEL"));

                            msg = WaitForMessage();

                            if (msg.Name.EndsWith(".No"))
                            {
                                proceed = true;
                            }

                        }
                        else
                        {
                            proceed = true;
                        }

                    }

                }

               



                //inform user
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    DefaultCultureInfo.Get(), BuildParamCollection(
                        "MessageKey", "MSG_RETURN_ADDED",
                        "NoTranslationOfMessageKey", false,
                        "ShowButton", true));
                WaitForMessage();



            }
            catch (Exception ex)
            {
                s_log.Error("", ex);
                if (!(ex is ProdumexException) || ((ProdumexException)ex).Error == null)
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "MessageKey", ex.Message,
                            "NoTranslationOfMessageKey", true,
                            "ShowButton", true));
                    WaitForMessage();

                }
                else
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "PmxError", ((ProdumexException)ex).Error,
                            "ShowButton", true));
                    WaitForMessage();
                }

                //retry?
                //ask user if he wants to retry
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                    DefaultCultureInfo.Get(), BuildParamCollection(
                        "TitleKey", "MSG_TITLE_RETRY_UPDATE",
                        "MessageKey", "MSG_QUESTION_AN_ERROR_OCCURED_WANT_RETRY"));
                msg = WaitForMessage();

                if (msg.Name.ToLower().EndsWith(".yes"))
                {
                    goto Step_AddReturn;
                }
            }

            //start new one
            goto Step_SelectCustomer;

            #endregion

        }

        #endregion

        #region HELPERS

        /// <summary>
        /// Removes the returned batches.
        /// </summary>
        /// <param name="salesReturnItemInfoList">The sales return item info list.</param>
        /// <param name="itemCode">The item code.</param>
        /// <param name="dsBatchNumbers">The ds batch numbers.</param>
        /// <returns></returns>
        private static DataSet RemoveReturnedBatches(Collection<SalesReturnItemInfo> salesReturnItemInfoList, string itemCode, DataSet dsBatchNumbers)
        {
            foreach (SalesReturnItemInfo returnInfo in salesReturnItemInfoList)
            {
                if (returnInfo.ItemCode == itemCode)
                {
                    for (int iLoop = 0; iLoop < dsBatchNumbers.Tables[0].Rows.Count; iLoop++)
                    {
                        DataRow row = dsBatchNumbers.Tables[0].Rows[iLoop];
                        double qtyToReduce = returnInfo.Quantity;
                        //does the batch number match?
                        if (returnInfo.BatchNumber == row[BatchNumberDataSet.Columns.BATCH_NUMBER.Name] as string)
                        {
                            //does the best before date match?
                            if ((!returnInfo.BestBeforeDate.HasValue &&
                                row[BatchNumberDataSet.Columns.BEST_BEFORE_DATE.Name] == System.DBNull.Value) ||
                                (returnInfo.BestBeforeDate.HasValue && row[BatchNumberDataSet.Columns.BEST_BEFORE_DATE.Name] is DateTime &&
                                returnInfo.BestBeforeDate.Value == Convert.ToDateTime(row[BatchNumberDataSet.Columns.BEST_BEFORE_DATE.Name])))
                            {
                                //reduce the quantity
                                double qtyOnBatch = Convert.ToDouble(row[BatchNumberDataSet.Columns.QUANTITY.Name]);
                                if (qtyOnBatch <= qtyToReduce)
                                {
                                    row.Delete();
                                    dsBatchNumbers.AcceptChanges();
                                    iLoop--;
                                    qtyToReduce -= qtyOnBatch;
                                    if (qtyToReduce <= 0)
                                    {
                                        break;
                                    }
                                }
                                else if (qtyOnBatch > qtyToReduce)
                                {
                                    qtyOnBatch -= qtyToReduce;
                                    row[BatchNumberDataSet.Columns.QUANTITY.Name] = qtyOnBatch;
                                    break;
                                }
                            }
                        }
                    }//iLoop
                }
            }//foreach 

            return dsBatchNumbers;
        }


        /// <summary>
        /// Executes the get PMX warehouse code.
        /// </summary>
        /// <param name="conn">The conn.</param>
        /// <param name="parameters">The parameters.</param>
        /// <returns></returns>
        private static PmxOseWarehouseInfo ExecuteGetPmxWarehouseCode(PmxDbConnection conn, object[] parameters)
        {
            PmxOseProvider oseProv = new PmxOseProvider(conn);
            return oseProv.GetPmxWarehouseInfo((string)parameters[0]);
        }
        
        ///// <summary>
        ///// Executes the print events.
        ///// </summary>
        ///// <param name="conn">The conn.</param>
        ///// <param name="parameters">The parameters.</param>
        ///// <returns></returns>
        //private static int ExecutePrintEvents(PmxDbConnection conn, object[] parameters)
        //{
        //    PmxPrintReportProvider prov = new PmxPrintReportProvider(conn);

        //    return prov.PrintReportsForEvent((PmxPrintReportEventType)parameters[0], (string)parameters[1], (string)parameters[2], (int)parameters[3]);
        //}


      
        /// <summary>
        /// Gets the number of rows.
        /// </summary>
        /// <param name="ds">The ds.</param>
        /// <returns></returns>
        private int GetNumberOfRows(DataSet ds)
        {
            if (ds != null &&
                ds.Tables.Count > 0)
            {
                return ds.Tables[0].Rows.Count;
            }

            return 0;
        }

        /// <summary>
        /// Checks if there are Products the on barcode.
        /// </summary>
        /// <param name="BarcodeContainerObject">The barcode container object.</param>
        /// <returns>
        /// True, if there is a product on the barcode, false if not
        /// </returns>
        public bool IsProductOnBarcode(BarcodeContainer BarcodeContainerObject)
        {
            if (BarcodeContainerObject != null &&
                !string.IsNullOrEmpty(BarcodeHelper.GetProductIdentification(BarcodeContainerObject)))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        /// <summary>
        /// Checks if there is a LU the on barcode.
        /// </summary>
        /// <param name="barcodeContainerObject">The barcode container object.</param>
        /// <returns>
        /// True, if there is a LU on the barcode, false if not
        /// </returns>
        public bool IsLUOnBarcode(BarcodeContainer barcodeContainerObject)
        {
            if (barcodeContainerObject != null &&
                barcodeContainerObject.GS1DataObject != null &&
                barcodeContainerObject.GS1DataObject.GTINLU != null)
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        /// <summary>
        /// Checks if there is an SSCC the on barcode.
        /// </summary>
        /// <param name="GS1DataScanned">The G s1 data scanned.</param>
        /// <returns>
        /// True, if there is an sscc on the barcode, false if not
        /// </returns>
        public bool IsSSCCOnBarcode(GS1Data GS1DataScanned)
        {
            if (GS1DataScanned != null &&
                (GS1DataScanned.Sscc != null))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        /// <summary>
        /// Checks if there is a best before Date the on barcode.
        /// </summary>
        /// <param name="GS1DataScanned">The G s1 data scanned.</param>
        /// <returns>
        /// True, if there is an best before date on the barcode, false if not
        /// </returns>
        public bool IsBestBeforeDateOnBarcode(GS1Data GS1DataScanned)
        {
            if (GS1DataScanned != null &&
                (GS1DataScanned.BestBeforeDate != null))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        /// <summary>
        /// Checks if there is a quantity the on barcode.
        /// </summary>
        /// <param name="GS1DataScanned">The G s1 data scanned.</param>
        /// <returns>
        /// True, if there is a quantity on the barcode, false if not
        /// </returns>
        public bool IsQuantityOnBarcode(BarcodeContainer barcode)
        {
            if (barcode != null &&
            barcode.GS1DataObject != null &&
                barcode.GS1DataObject.Count != null)
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        /// <summary>
        /// Checks if there is a batch number the on barcode.
        /// </summary>
        /// <param name="GS1DataScanned">The G s1 data scanned.</param>
        /// <returns>
        /// True, if there is a batch number on the barcode, false if not
        /// </returns>
        public bool IsBatchNumberOnBarcode(GS1Data GS1DataScanned)
        {
            if (GS1DataScanned != null &&
                (GS1DataScanned.LotNumber != null))
            {
                return true;
            }
            else
            {
                return false;
            }
        }


        /// <summary>
        /// Executes the get item info.
        /// </summary>
        /// <param name="conn">The conn.</param>
        /// <param name="parameters">The parameters.</param>
        /// <returns></returns>
        private static PmxItemInfo ExecuteGetItemInfo(PmxDbConnection conn, object[] parameters)
        {
            PmxItemAllConnectionsProvider prov = new PmxItemAllConnectionsProvider(conn);
            return prov.GetItemInfo((string)parameters[0]);
        }





        #endregion

        #region BUILD QUERY

        /// <summary>
        /// Static class grouping our native queries
        /// </summary>
        internal static class BuildQuery
        {
            private static SboBooleanConvertor m_sboBooleanConvertor = null;
            private static SboBooleanConvertor BooleanConvertor
            {
                get
                {
                    if (m_sboBooleanConvertor == null)
                    {
                        m_sboBooleanConvertor = new SboBooleanConvertor();
                    }
                    return m_sboBooleanConvertor;
                }
            }

            /// <summary>
            /// Gets the zone types for items.
            /// </summary>
            /// <param name="itemCodes">The item codes.</param>
            /// <returns></returns>
            public static string GetZoneTypesForItems(Collection<string> itemCodes)
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT DISTINCT");

                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxItemZoneTypeTable.NAME, PmxItemZoneTypeTable.Columns.ZoneTypeCode.NAME, PmxItemZoneTypeTable.Columns.ZoneTypeCode.NAME);

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxItemZoneTypeTable.NAME));


                string operand = "";
                strBuilder.Append(" WHERE ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxItemZoneTypeTable.NAME, PmxItemZoneTypeTable.Columns.ItemCode.NAME);
                strBuilder.Append(" IN ( ");

                foreach (string itemCode in itemCodes)
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, CultureInfo.InvariantCulture));
                    operand = ", ";
                }
                strBuilder.Append(") ");


                return strBuilder.ToString();
            }

            internal static string GetSSCCInfo(string sscc)
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat("SELECT DISTINCT {0}",
                    SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.InternalKey.NAME, SSCCDataSet.Columns.LUID.Name));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.SSCC.NAME, SSCCDataSet.Columns.SSCC.Name));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME, SSCCDataSet.Columns.STORAGE_LOCATION_CODE.Name));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME, SSCCDataSet.Columns.PMX_WAREHOUSE_CODE.Name));

                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Name.NAME, SSCCDataSet.Columns.PMX_WAREHOUSE_NAME.Name));


                strBuilder.AppendFormat("\r\n FROM {0}",
                    SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));

                strBuilder.AppendFormat("\r\n INNER JOIN {0}",
                    SqlCommandHelper.EscapeTableName(PmxLogisticUnitIDTable.NAME));

                strBuilder.AppendFormat(" ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.InternalKey.NAME),
                    SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME));

                strBuilder.AppendFormat("\r\n INNER JOIN {0}",
                SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));

                strBuilder.AppendFormat(" ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME),
                    SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME));

                strBuilder.AppendFormat("\r\n INNER JOIN {0}",
                    SqlCommandHelper.EscapeTableName(PmxOseWarehouseTable.NAME));
                strBuilder.AppendFormat(" ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME),
                    SqlCommandHelper.EscapeTableColumnName(PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Code.NAME));




                strBuilder.AppendFormat("\r\n WHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.SSCC.NAME),
                    SqlCommandHelper.FormatAndEscapeColumnValue(sscc, CultureInfo.InvariantCulture));

                return strBuilder.ToString();
            }

            internal static string GetZoneTypesForItem(string itemCode)
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat("SELECT DISTINCT {0}",
                    SqlCommandHelper.EscapeTableColumnName(PmxItemZoneTypeTable.NAME, PmxItemZoneTypeTable.Columns.ZoneTypeCode.NAME));

                strBuilder.AppendFormat(" FROM {0}",
                    SqlCommandHelper.EscapeTableName(PmxItemZoneTypeTable.NAME));

                strBuilder.AppendFormat(" WHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxItemZoneTypeTable.NAME, PmxItemZoneTypeTable.Columns.ItemCode.NAME),
                    SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, CultureInfo.InvariantCulture));

                return strBuilder.ToString();
            }


            /// <summary>
            /// Gets the name and the code of the  location, ordered by name.
            /// </summary>
            /// <param name="locationCode">The location code.</param>
            /// <param name="warehouseCode">The warehouse code.</param>
            /// <param name="zoneTypeCodes">The zone type codes.</param>
            /// <param name="emptyLocations">if set to <c>true</c> [empty locations].</param>
            /// <returns>The sql statement to get the location</returns>
            internal static string GetStorageLocationForZoneTypes(string locationCode, string warehouseCode, IList<string> zoneTypeCodes, bool emptyLocations)
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT ");

                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.Code.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(LocationDataSet.Columns.CODE.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.Name.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(LocationDataSet.Columns.DESCRIPTION.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(LocationDataSet.Columns.WAREHOUSECODE.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Name.NAME, LocationDataSet.Columns.WAREHOUSENAME.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.BlockMoveWhenNotEmpty.NAME));


                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);


                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseWarehouseTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Code.NAME);


                if (emptyLocations)
                {
                    strBuilder.Append(" LEFT JOIN ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(" ON ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME);
                    strBuilder.Append(" = ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);

                }

                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.IsActive.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));


                if (!string.IsNullOrEmpty(locationCode))
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.Code.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(locationCode, CultureInfo.InvariantCulture));
                }

                if (warehouseCode != null &&
                    warehouseCode.Trim() != "")
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(warehouseCode, CultureInfo.InvariantCulture));

                }

                if (zoneTypeCodes != null &&
                    zoneTypeCodes.Count > 0)
                {
                    foreach (string zoneTypeCode in zoneTypeCodes)
                    {
                        strBuilder.Append(" AND ");
                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                        strBuilder.Append(".");
                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.Code.NAME));
                        strBuilder.Append(" IN ( ");

                        strBuilder.Append("SELECT ");

                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                        strBuilder.Append(".");
                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.Code.NAME));


                        strBuilder.Append(" FROM ");
                        strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseStorageLocationGeneralTable.NAME));
                        strBuilder.Append(" INNER JOIN ");
                        strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                        strBuilder.Append(" ON ");
                        SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                        strBuilder.Append(" = ");
                        SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);
                        strBuilder.Append(" RIGHT OUTER JOIN ");
                        strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxLinkZoneZoneTypeTable.NAME));
                        strBuilder.Append(" ON ");
                        SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxZoneCode.NAME);
                        strBuilder.Append(" = ");
                        SqlCommandHelper.AddTableColumnName(strBuilder, PmxLinkZoneZoneTypeTable.NAME, PmxLinkZoneZoneTypeTable.Columns.ParentCode.NAME);


                        strBuilder.Append(" WHERE ");


                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxLinkZoneZoneTypeTable.NAME));
                        strBuilder.Append(".");
                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxLinkZoneZoneTypeTable.Columns.ZoneTypeCode.NAME));
                        strBuilder.Append(" = ");
                        strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(zoneTypeCode, CultureInfo.InvariantCulture));


                        strBuilder.Append(") ");

                    }
                }

                if (emptyLocations)
                {
                    //If there is no link to the inventory, the location is empty
                    strBuilder.Append(" AND ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME);
                    strBuilder.Append(" IS NULL ");

                }

                strBuilder.Append(" ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.Name.NAME));


                return strBuilder.ToString();
            }


            /// <summary>
            /// Command for getting all logistic carriers order by name
            /// </summary>
            /// <param name="onlyInStock">if set to <c>true</c> the system will get the logistic carriers that are in stock and have no products on it (LUID IS NULL)</param>
            /// <param name="filterOutFrozenItems">if set to <c>true</c> [filter out frozen items].</param>
            /// <returns>
            /// Command for getting all logistic carriers order by name
            /// </returns>
            public static string GetLogisticCarriersOrderByName(bool onlyInStock, bool filterOutFrozenItems, DbTool dbTool)
            {
				DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();

                strBuilder.Append("SELECT DISTINCT ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.ItemCode.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(ProductDataSet.Columns.PRODUCT_CODE.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.CustomItemDescr.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(ProductDataSet.Columns.PRODUCT_DESCRIPTION.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.CodeBars.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(ProductDataSet.Columns.GTIN.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.IsLogisticCarrier.NAME));

                strBuilder.Append("\r\n FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboItemTable.NAME));

                if (onlyInStock)
                {
                    strBuilder.Append("\r\n INNER JOIN ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(" ON ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.ItemCode.NAME);
                    strBuilder.Append(" = ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME);
                    strBuilder.Append(" AND ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME);
                    strBuilder.Append(" IS NULL ");
                }

                strBuilder.Append("\r\n WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.InventoryItem.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.IsLogisticCarrier.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

                if (filterOutFrozenItems)
                {
                    Produmex.Sbo.Logex.Data.SboCommands.GetFrozenItemFilterSQLWherePart(strBuilder, " AND ");
                }

                strBuilder.Append("\r\n ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.CustomItemDescr.NAME));

                return strBuilder.ToString();
            }

            /// <summary>
            /// Gets the sales retour reasons.
            /// </summary>
            /// <returns></returns>
            public static string GetSalesRetourReasons()
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT  ");

                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxReasonTable.NAME, PmxReasonTable.Columns.Code.NAME, ReasonDataSet.Columns.REASON_CODE.Name);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxReasonTable.NAME, PmxReasonTable.Columns.Name.NAME, ReasonDataSet.Columns.REASON_NAME.Name);

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxReasonTable.NAME));

                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxReasonTable.Columns.UseForSalesRetour.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

                return strBuilder.ToString();
            }


            /// <summary>
            /// Command for getting the open quantity
            /// </summary>
            /// <param name="salesDeliveryNoteDocEntry">The sales delivery note doc entry.</param>
            /// <param name="itemCode">The item code.</param>
            /// <param name="batchNumber">The batch number.</param>
            /// <param name="bestBeforeDate">The best before date.</param>
            /// <param name="dbCultureInfo">The culture info.</param>
            /// <returns>Command for getting the open quantity</returns>
            public static string GetQueryGetReturnedQtyFromSalesDelivery(int salesDeliveryNoteDocEntry, string itemCode, string batchNumber, DateTime? bestBeforeDate)
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT SUM(");

                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.Quantity.NAME);
                strBuilder.Append(") AS ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryDetailTable.Columns.Quantity.NAME));

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxSalesReturnLineTable.NAME));

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryDetailTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.DocEntry.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxSalesReturnLineTable.NAME, PmxSalesReturnLineTable.Columns.DocEntry.NAME);
                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.TransType.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxSalesReturnLineTable.NAME, PmxSalesReturnLineTable.Columns.ObjectType.NAME);
                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.DocLineNum.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxSalesReturnLineTable.NAME, PmxSalesReturnLineTable.Columns.LineNum.NAME);

                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxItemTransactionalInfoTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.ItemTransactionalInfoKey.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.InternalKey.NAME);




                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxSalesReturnLineTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxSalesReturnLineTable.Columns.BaseEntry.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(salesDeliveryNoteDocEntry, CultureInfo.InvariantCulture));

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryDetailTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryDetailTable.Columns.TransType.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(((int)SboLinkDocumentType.SalesReturn).ToString(), CultureInfo.InvariantCulture));

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryDetailTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryDetailTable.Columns.ItemCode.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, CultureInfo.InvariantCulture));

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.Columns.BatchNumber1.NAME));

                if (!string.IsNullOrEmpty(batchNumber))
                {
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(batchNumber, CultureInfo.InvariantCulture));
                }
                else
                {
                    strBuilder.Append(" IS NULL ");
                }

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.Columns.BestBeforeDate.NAME));

                if (bestBeforeDate.HasValue)
                {
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(bestBeforeDate.Value, CultureInfo.InvariantCulture));
                }
                else
                {
                    strBuilder.Append(" IS NULL ");
                }

                strBuilder.Append(" GROUP BY ");

                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryDetailTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryDetailTable.Columns.ItemCode.NAME));

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxItemTransactionalInfoTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.Columns.BatchNumber1.NAME));

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxItemTransactionalInfoTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.Columns.BestBeforeDate.NAME));

                return strBuilder.ToString();
            }

            /// <summary>
            /// Gets the batch numbers.
            /// </summary>
            /// <param name="salesDeliveryNoteDocEntry">The sales delivery note doc entry.</param>
            /// <param name="itemCode">The item code.</param>
            /// <param name="batchNumber">The batch number.</param>
            /// <param name="bestBeforeDate">The best before date.</param>
            /// <param name="sscc">The SSCC.</param>
            /// <returns></returns>
            public static string GetBatchNumbersForSalesRetour(int salesDeliveryNoteDocEntry, string itemCode, string batchNumber, DateTime? bestBeforeDate, string sscc)
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT ");

                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BestBeforeDate.NAME, BatchNumberDataSet.Columns.BEST_BEFORE_DATE.Name));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BatchNumber1.NAME, BatchNumberDataSet.Columns.BATCH_NUMBER.Name));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BatchNumber2.NAME, BatchNumberDataSet.Columns.BATCH_NUMBER2.Name));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.ItemTransactionalInfoKey.NAME, BatchNumberDataSet.Columns.ITEMTRANSACTIONALINFOKEY.Name));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.QualityStatusCode.NAME, BatchNumberDataSet.Columns.QUALITY_STATUS_CODE.Name));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.Uom.NAME, PmxInventoryDetailTable.Columns.Uom.NAME));
                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.QuantityPerUom.NAME, PmxInventoryDetailTable.Columns.QuantityPerUom.NAME));
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboItemTable.NAME, SboItemTable.Columns.InventoryUom.NAME);

                strBuilder.Append(", SUM (ISNULL(");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.Quantity.NAME);
                strBuilder.Append(", 0) * -1) AS ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(BatchNumberDataSet.Columns.QUANTITY.Name));


                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryDetailTable.NAME));
                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboItemTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.ItemCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME);

                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxItemTransactionalInfoTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.ItemTransactionalInfoKey.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.InternalKey.NAME);


                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryDetailTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryDetailTable.Columns.DocEntry.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(salesDeliveryNoteDocEntry, CultureInfo.InvariantCulture));

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryDetailTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryDetailTable.Columns.TransType.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(((int)SboLinkDocumentType.SalesDeliveryNote).ToString(), CultureInfo.InvariantCulture));

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryDetailTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryDetailTable.Columns.ItemCode.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, CultureInfo.InvariantCulture));

                if (!string.IsNullOrEmpty(batchNumber))
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.Columns.BatchNumber1.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(batchNumber, CultureInfo.InvariantCulture));
                }

                if (bestBeforeDate.HasValue)
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTransactionalInfoTable.Columns.BestBeforeDate.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(bestBeforeDate.Value, CultureInfo.InvariantCulture));
                }

                if (!string.IsNullOrEmpty(sscc))
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryDetailTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryDetailTable.Columns.SSCC.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(sscc, CultureInfo.InvariantCulture));
                }

                strBuilder.Append(" GROUP BY ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BestBeforeDate.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BatchNumber1.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxItemTransactionalInfoTable.NAME, PmxItemTransactionalInfoTable.Columns.BatchNumber2.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.ItemTransactionalInfoKey.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.DocEntry.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.ItemCode.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.QualityStatusCode.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.Uom.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.QuantityPerUom.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboItemTable.NAME, SboItemTable.Columns.InventoryUom.NAME);


                return strBuilder.ToString();
            }



			/// <summary>
			/// Command for getting all products order by name
			/// </summary>
			/// <param name="gtin">The gtin.</param>
			/// <param name="sscc">The SSCC.</param>
			/// <param name="salesDeliveryNoteDocEntry">The sales delivery note doc entry.</param>
			/// <param name="dbTool">The database tool.</param>
			/// <returns>
			/// Command for getting all products order by name
			/// </returns>
            public static string GetProductForSalesDeliveryNoteOrderByName(string gtin, string sscc, int? salesDeliveryNoteDocEntry, DbTool dbTool)
            {
				DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();
                strBuilder.Append("SELECT DISTINCT ");

                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.ItemCode.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(ProductDataSet.Columns.PRODUCT_CODE.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.CustomItemDescr.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(ProductDataSet.Columns.PRODUCT_DESCRIPTION.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.CodeBars.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(ProductDataSet.Columns.GTIN.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.HasBestBeforeDate.NAME));

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.ManBtchNum.NAME));

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.IsLogisticUnit.NAME));

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboItemTable.NAME, SboItemTable.Columns.SalesUom.NAME);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboItemTable.NAME, SboItemTable.Columns.InventoryUom.NAME);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboItemTable.NAME, SboItemTable.Columns.NumberOfItemsPerSalesUnit.NAME);

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxItemTable.NAME));

                if (salesDeliveryNoteDocEntry.HasValue)
                {
                    //inner join Product - sales delivery line Line
                    strBuilder.Append(" INNER JOIN ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxSalesDeliveryNoteLineTable.NAME));
                    strBuilder.Append(" ON ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxSalesDeliveryNoteLineTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxSalesDeliveryNoteLineTable.Columns.ItemCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxItemTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxItemTable.Columns.ItemCode.NAME));

                    //inner join Sales Delivery Note - Sales Delivery Note Line
                    strBuilder.Append(" INNER JOIN ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxSalesDeliveryNoteTable.NAME));
                    strBuilder.Append(" ON ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxSalesDeliveryNoteTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxSalesDeliveryNoteTable.Columns.DocEntry.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxSalesDeliveryNoteLineTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxSalesDeliveryNoteLineTable.Columns.DocEntry.NAME));

                    strBuilder.Append(" INNER JOIN ");

                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryDetailTable.NAME));
                    strBuilder.Append(" ON ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.DocEntry.NAME);
                    strBuilder.Append(" = ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxSalesDeliveryNoteLineTable.NAME, PmxSalesDeliveryNoteLineTable.Columns.DocEntry.NAME);
                    strBuilder.Append("AND ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.TransType.NAME);
                    strBuilder.Append(" = ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxSalesDeliveryNoteLineTable.NAME, PmxSalesDeliveryNoteLineTable.Columns.ObjectType.NAME);
                    strBuilder.Append("AND ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryDetailTable.NAME, PmxInventoryDetailTable.Columns.DocLineNum.NAME);
                    strBuilder.Append(" = ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxSalesDeliveryNoteLineTable.NAME, PmxSalesDeliveryNoteLineTable.Columns.LineNum.NAME);

                }

                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxItemPackagingTypeTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxItemPackagingTypeTable.NAME, PmxItemPackagingTypeTable.Columns.ItemCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME);


                string operand = " WHERE ";

                strBuilder.Append(operand);
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.InventoryItem.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));
                operand = " AND ";

                if (!string.IsNullOrEmpty(gtin))
                {
                    strBuilder.Append(operand);
                    Produmex.Sbo.Logex.Data.SboCommands.AddWhereClauseSearchForBarcode(strBuilder, gtin, true);
                    operand = " AND ";
                }

                if (salesDeliveryNoteDocEntry.HasValue)
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxSalesDeliveryNoteTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxSalesDeliveryNoteTable.Columns.DocEntry.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(salesDeliveryNoteDocEntry.Value, CultureInfo.InvariantCulture));
                    operand = " AND ";

                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxSalesDeliveryNoteLineTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxSalesDeliveryNoteLineTable.Columns.OpenQty.NAME));
                    strBuilder.Append(" > ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(0, CultureInfo.InvariantCulture));
                    operand = " AND ";

                    if (!string.IsNullOrEmpty(sscc))
                    {
                        strBuilder.Append(operand);
                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryDetailTable.NAME));
                        strBuilder.Append(".");
                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxInventoryDetailTable.Columns.SSCC.NAME));
                        strBuilder.Append(" = ");
                        strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(sscc, CultureInfo.InvariantCulture));
                        operand = " AND ";
                    }
                }


                strBuilder.Append(operand);
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.SalesItem.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));
                operand = " AND ";


                Logex.Data.SboCommands.GetFrozenItemFilterSQLWherePart(strBuilder, operand);

                strBuilder.Append(" ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.CustomItemDescr.NAME));
                return strBuilder.ToString();
            }

            /// <summary>
            /// Command for getting all Sales delivery notes order by date
            /// </summary>
            /// <param name="cardCode">The card code.</param>
            /// <returns>
            /// Command for getting all Sales delivery notes by date
            /// </returns>
            public static string GetSalesDeliveryNotesOrderByDate(string cardCode)
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT ");

                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboSalesDeliveryNoteTable.Columns.CardCode.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(SalesDeliveryNoteDataSet.Columns.CUSTOMER_CODE.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboSalesDeliveryNoteTable.Columns.CardName.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(SalesDeliveryNoteDataSet.Columns.CUSTOMER_NAME.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboSalesDeliveryNoteTable.Columns.DocDueDate.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(SalesDeliveryNoteDataSet.Columns.DOC_DUE_DATE.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboSalesDeliveryNoteTable.Columns.DocEntry.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(SalesDeliveryNoteDataSet.Columns.DOC_ENTRY.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboSalesDeliveryNoteTable.Columns.DocNum.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(SalesDeliveryNoteDataSet.Columns.DOC_NUM.Name);

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboSalesDeliveryNoteTable.NAME));

                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboSalesDeliveryNoteTable.Columns.DocStatus.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboDocStatusColumnValues.Open.VALUE, CultureInfo.InvariantCulture));

                if (!string.IsNullOrEmpty(cardCode))
                {

                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboSalesDeliveryNoteTable.Columns.CardCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(cardCode, CultureInfo.InvariantCulture));
                }

                strBuilder.Append(" ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboSalesDeliveryNoteTable.Columns.DocDueDate.NAME));
                strBuilder.Append(" DESC ");
                return strBuilder.ToString();
            }



            /// <summary>
            /// Command for getting all customers order by name
            /// </summary>
            /// <returns>Command for getting all customers order by name</returns>
            public static string GetCodeNameCustomersOrderByName()
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT ");

                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboBusinessPartnerTable.Columns.CardCode.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(CustomerDataSet.Columns.CUSTOMER_CODE.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboBusinessPartnerTable.Columns.CardName.NAME));
                strBuilder.Append(" AS ");
                strBuilder.Append(CustomerDataSet.Columns.CUSTOMER_NAME.Name);

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboBusinessPartnerTable.NAME));

                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboBusinessPartnerTable.Columns.CardType.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBusinessPartnerTypeColumnValues.Customer.VALUE, CultureInfo.InvariantCulture));

                strBuilder.Append(" ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboBusinessPartnerTable.Columns.CardName.NAME));
                return strBuilder.ToString();
            }


           




        }

        #endregion

    }


}
